# Module containing the different decorators used
import functools
import time


def timer(func):
    """
    Used to decorate a function to print the time spent running it
    """

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer


def timer_min(func):
    """
    Used to decorate a function to print the time spent running it
    """

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        print(f"Finished {func.__name__!r} in {run_time / 60.0:.4f} mins")
        return value

    return wrapper_timer



import unittest
from pydelling.preprocessing.gid_preprocessor import *


class GidPreprocessorTest(unittest.TestCase):
    def test_create_line(self):
        point_1 = Point([0.0, 0.0, 0.0])
        point_2 = Point([1.0, 0.0, 0.0])
        line = Line(point_1, point_2)
        class TestGidobject(GidObject):
            def construct(self, *args, **kwargs):
                self.add([point_1, point_2])
                self.add(line)

        test_gid = TestGidobject()
        test_gid.run(internal=True)

    def test_polyline(self):
        point_1 = Point([0.0, 0.0, 0.0])
        point_2 = Point([1.0, 0.0, 0.0])
        point_3 = Point([1.0, 1.0, 0.0])
        point_4 = Point([0.0, 1.0, 0.0])
        polyline = Polyline([point_1, point_2, point_3, point_4], connect=True)
        polyline.run()

    def test_surface(self):
        point_1 = Point([0.0, 0.0, 0.0])
        point_2 = Point([1.0, 0.0, 0.0])
        point_3 = Point([1.0, 1.0, 0.0])
        point_4 = Point([0.0, 1.0, 0.0])

        polyline = Polyline([point_1, point_2, point_3, point_4], connect=True)
        surface = Surface(polyline.lines)

        class TestSurface(GidObject):
            def construct(self, *args, **kwargs):
                self.add(polyline)
                self.add(surface)

        test_surface = TestSurface()
        test_surface.run()

    def test_overlapping_points(self):
        point_1 = Point([0.0, 0.0, 0.0])
        point_2 = Point([0.0, 0.0, 0.0])
        point_3 = Point([1.0, 0.0, 0.0])
        point_4 = Point([2.0, 0.0, 0.0])
        point_5 = Point([1.0, 0.0, 0.0])
        point_6 = Point([2.0, 0.0, 0.0])

        class TestPoints(GidObject):
            def construct(self, *args, **kwargs):
                self.add(point_1)
                self.add(point_2)
                self.add(point_3)
                self.add(point_4)
                self.add(point_5)
                self.add(point_6)
                line_1 = Line(point_1, point_6)
                self.add(line_1)

        test_points = TestPoints()
        test_points.run()

    def test_double_surface(self):
        point_1 = Point([0.0, 0.0, 0.0])
        point_2 = Point([1.0, 0.0, 0.0])
        point_3 = Point([1.0, 1.0, 0.0])
        point_4 = Point([0.0, 1.0, 0.0])
        point_5 = Point([0.0, 2.0, 0.0])

        polyline = Polyline([point_1, point_2, point_3, point_4], connect=True)

        polyline_2 = Polyline([point_1, point_2, point_3, point_5], connect=True)

        surface = Surface(polyline.lines)
        surface_2 = Surface(polyline_2.lines)
        class TestSurface(GidObject):
            def construct(self, *args, **kwargs):
                self.add(polyline)
                self.add(polyline_2)

                self.add(surface)
                self.add(surface_2)

        test_surface = TestSurface()
        test_surface.run()

    def test_lines_equal(self):
        point_1 = Point([0.0, 0.0, 0.0])
        point_2 = Point([1.0, 0.0, 0.0])
        point_3 = Point([1.0, 1.0, 0.0])

        line_1 = Line(point_1, point_2)
        line_2 = Line(point_1, point_2)
        line_3 = Line(point_1, point_3)
        line_4 = Line(point_2, point_1)

        self.assertTrue(Line.check_lines_equal(line_1, line_2))
        self.assertFalse(Line.check_lines_equal(line_1, line_3))
        self.assertTrue(Line.check_lines_equal(line_1, line_4))

    def tearDown(self) -> None:
        from pathlib import Path
        gid_file = Path("gid_batch.bch")
        if gid_file.exists():
            gid_file.unlink()


if __name__ == '__main__':
    unittest.main()


import logging
import numpy as np
import functools
import scipy

logger = logging.getLogger(__name__)


class UnitConverter:
    """A class for converting between units of measurement."""
    def convert_time(self, value, initial_unit, final_unit):
        """Converts between time units."""
        value = float(value)
        if initial_unit == 's' and final_unit == 'd':
            return value / 86400
        elif initial_unit == 'd' and final_unit == 's':
            return value * 86400
        elif initial_unit == 's' and final_unit == 'y':
            return value / 31557600
        elif initial_unit == 'y' and final_unit == 's':
            return value * 31557600
        elif initial_unit == 'd' and final_unit == 'y':
            return value / 365
        elif initial_unit == 'y' and final_unit == 'd':
            return value * 365
        elif initial_unit == 'min' and final_unit == 'y':
            return value / 525600
        elif initial_unit == 'y' and final_unit == 'min':
            return value * 525600
        elif initial_unit == 'min' and final_unit == 'd':
            return value / 1440
        elif initial_unit == 'd' and final_unit == 'min':
            return value * 1440
        elif initial_unit == 'min' and final_unit == 's':
            return value / 60
        elif initial_unit == 's' and final_unit == 'min':
            return value * 60
        elif initial_unit == 'min' and final_unit == 'h':
            return value / 60
        elif initial_unit == 'h' and final_unit == 'min':
            return value * 60
        elif initial_unit == 'h' and final_unit == 'd':
            return value / 24
        elif initial_unit == 'd' and final_unit == 'h':
            return value * 24
        elif initial_unit == 'h' and final_unit == 's':
            return value * 3600
        elif initial_unit == 's' and final_unit == 'h':
            return value / 3600
        elif initial_unit == 'h' and final_unit == 'y':
            return value / 8760
        elif initial_unit == 'y' and final_unit == 'h':
            return value * 8760
        elif initial_unit == final_unit:
            return value
        else:
            raise ValueError(f"Invalid unit conversion: {initial_unit} to {final_unit}")

class SemistructuredFinder:
    """A class for processing a semi-structured grid based on a set of points.
    Based on a cloud of points, the algorithm creates clusters of points that share a common z coordinate.
    """
    engine = None  # The clustering engine used to generate the clusters
    is_run = False
    projection_dict = {'x': 0, 'y': 1, 'z': 2}
    def __init__(self,
                 points,
                 n_clusters=None,
                 eps=0.1,
                 min_samples=10,
                 projection_axis: str = 'z',
                 **kwargs,
                 ):
        """Initializes the class.
        Args:
            points (np.array): A numpy array of points.
            n_clusters (int): The number of clusters to find. If None, the number of clusters is determined automatically.
            eps (float): The maximum distance between two samples for one to be considered as in the neighborhood of the other.
            min_samples (int): The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.
            projection_axis (str): The axis along which to project the points. The axis is assumed to be the third dimension.
            **kwargs: Additional keyword arguments.
        """

        self.points = points  # A list of points to consider
        # Project the points onto the projection axis
        self.projected_points = []
        self.projection_axis = projection_axis
        for i in range(0, 3):
           if i != self.projection_dict[projection_axis]:
                self.projected_points.append(self.points[:, i])
        self.projected_points = np.array(self.projected_points).T

        self._n_clusters = n_clusters  # The number of clusters to generate
        if self._n_clusters is None:
            logger.info(f"Generating clusters of {len(self.projected_points)} points using the DBSCAN algorithm with eps={eps} and min_samples={min_samples}")
            self.clusters = self.generate_clusters_dbscan(eps=eps, min_samples=min_samples, **kwargs)
        else:
            logger.info(f"Generating clusters of {len(self.projected_points)} points using the KMeans algorithm with n_clusters={n_clusters}")
            self.clusters = self.generate_clusters_kmeans(n_clusters=n_clusters, **kwargs)
        self.is_run = True

    def generate_clusters_dbscan(self, eps, min_samples, **kwargs):
        """Generates clusters of points using the DBSCAN algorithm."""
        import numpy as np
        from sklearn.cluster import DBSCAN
        from sklearn.preprocessing import StandardScaler
        # Standardize the data

        # Compute DBSCAN
        self.engine: DBSCAN = DBSCAN(eps=eps, min_samples=min_samples, **kwargs)
        self.engine.fit(self.projected_points)
        core_samples_mask = np.zeros_like(self.engine.labels_, dtype=bool)
        core_samples_mask[self.engine.core_sample_indices_] = True
        labels = self.engine.labels_

        # Number of clusters in labels, ignoring noise if present.
        n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)

        # Create a list of clusters
        clusters = []
        for i in range(n_clusters_):
            clusters.append([])

        # Add points to clusters
        for i, label in enumerate(labels):
            if label != -1:
                clusters[label].append(self.points[i])

        return clusters

    def generate_clusters_kmeans(self, n_clusters, **kwargs):
        """Generates clusters of points using the KMeans algorithm."""
        from sklearn.cluster import KMeans
        self.engine: KMeans = KMeans(n_clusters=self._n_clusters, random_state=0, **kwargs)
        self.engine.fit(self.projected_points)
        labels = self.engine.labels_
        clusters = []
        for i in range(n_clusters):
            clusters.append([])
        for i, label in enumerate(labels):
            clusters[label].append(self.points[i])
        return clusters

    @functools.cached_property
    def n_clusters(self):
        """Returns the number of clusters."""
        return len(self.clusters)

    @functools.cached_property
    def point_idx(self):
        """Returns a list of indices of the points in the clusters."""
        point_idx = []
        for i in range(self.n_clusters):
            point_idx.append([])
        for i, label in enumerate(self.engine.labels_):
            if label != -1:
                point_idx[label].append(i)
        return point_idx

    def get_point_idx_in_cluster(self, cluster_idx):
        """Returns the indices of the points in the specified cluster."""
        return self.point_idx[cluster_idx]

    def get_closest_cluster_from_point(self, point: np.array):
        """Returns the cluster closest to a point."""
        dist = []
        # Compute the projected point
        projected_point = []
        for i in range(0, 3):
            if i != self.projection_dict[self.projection_axis]:
                projected_point.append(point[i])
        projected_point = np.array(projected_point)
        for cluster_center in self.projected_cluster_centers:
            dist.append(np.linalg.norm(projected_point - cluster_center))
        return self.clusters[np.argmin(dist)]

    def get_closest_cluster_from_xy(self, x: float, y: float):
        """Returns the cluster closest to a point."""
        return self.get_closest_cluster_from_point(np.array([x, y, 0]))

    def get_closest_point_ids_from_point(self, point: np.array):
        """Returns the cluster closest to a point using KDTree."""
        import numpy as np
        from scipy.spatial import KDTree
        projected_point = []
        for i in range(0, 3):
            if i != self.projection_dict[self.projection_axis]:
                projected_point.append(point[i])
        projected_point = np.array(projected_point)
        dist, idx = self.projected_cluster_centers_kdtree.query(projected_point)
        return self.point_idx[idx]

    def get_closest_point_ids_from_xy(self, x: float, y: float):
        """Returns the cluster closest to a point."""
        return self.get_closest_point_ids_from_point(np.array([x, y, 0]))

    @functools.cached_property
    def projected_cluster_centers(self):
        """Returns the cluster centers."""

        centers = []
        for cluster in self.clusters:
            projected_points = []
            for cluster_point in cluster:
                projected_points.append([])
                for i in range(0, 3):
                    if i != self.projection_dict[self.projection_axis]:
                        projected_points[-1].append(cluster_point[i])
            projected_points = np.array(projected_points)
            centers.append(np.mean(projected_points, axis=0))
        return centers

    @functools.cached_property
    def projected_cluster_centers_kdtree(self) -> scipy.spatial.KDTree:
        """Returns the cluster centers."""
        return scipy.spatial.KDTree(self.projected_cluster_centers)



    @functools.cached_property
    def cluster_centers(self):
        """Returns the cluster centers."""
        centers = []
        for cluster in self.clusters:
            centers.append(np.mean(cluster, axis=0))
        return centers










from __future__ import annotations

from typing import List

import numpy as np

from .BasePrimitive import BasePrimitive
from .Point import Point


class Vector(np.ndarray, BasePrimitive):
    coords: np.ndarray
    def __new__(cls, v: np.ndarray or List=None, p1: Point or List or np.ndarray=None, p2: Point or List or np.ndarray=None, *args, **kwargs):
        if v is not None:
            assert len(v) >= 2, "Vector must have 2 or 3 coordinates"
            if isinstance(v, Vector):
                obj = np.asarray(v).view(cls)
            else:
                obj = np.asarray(v).view(cls)
            return obj


        elif p1 is not None and p2 is not None:
            p1 = Point(p1)
            p2 = Point(p2)
            obj = np.asarray(p2 - p1).view(cls)
            return obj

        else:
            raise ValueError("Vector must be initialized with either a list or a numpy array")

    def __repr__(self):
        return f"Vector({self})"




import numpy as np

from pydelling.readers import RasterFileReader
from stl import mesh
import logging
from pydelling.utils import create_results_folder

logger = logging.getLogger(__name__)

class ClosedStlGenerator(object):
    """This class generates a closed STL from two regular raster files"""
    def __init__(self, bottom_surface: RasterFileReader,
                 top_surface: RasterFileReader = None,
                 aperture: RasterFileReader = None,
    ):
        # Generate needed data
        self.bottom_surface = bottom_surface
        if top_surface is not None:
            self.top_surface = top_surface
        if aperture is not None:
            self.aperture = aperture
        else:
            self.aperture = top_surface - bottom_surface
        assert self.aperture is not None, "Aperture is not defined"
        self.stl_file: mesh.Mesh

    def run(self, output_filename: str = 'closed_stl.stl',
            export_faces=True,
            ):
        """This method runs the closed STL generator"""
        logger.info(f'Generating closed STL based on bottom raster file: {self.bottom_surface.filename} and top raster file: {self.top_surface.filename}')
        # First, we need to generate the vertices
        vertices = []
        # Now, we need to generate the faces
        faces = []
        # First, we need to generate the faces for the bottom surface
        logger.info('Generating faces for the bottom surface')
        faces_dict = {
            'bottom': [],
            'top': [],
            'west': [],
            'east': [],
            'north': [],
            'south': [],
        }
        for row in range(self.bottom_surface.nrows - 1):
            for col in range(self.bottom_surface.ncols - 1):
                # Generate the face
                # Save the face vertices
                vertex_1 = [self.bottom_surface.x_mesh[row, col], self.bottom_surface.y_mesh[row, col], self.bottom_surface.data[row, col]]
                vertex_2 = [self.bottom_surface.x_mesh[row, col + 1], self.bottom_surface.y_mesh[row, col + 1], self.bottom_surface.data[row, col + 1]]
                vertex_3 = [self.bottom_surface.x_mesh[row + 1, col + 1], self.bottom_surface.y_mesh[row + 1, col + 1], self.bottom_surface.data[row + 1, col + 1]]
                face_vertices = [vertex_1, vertex_2, vertex_3]
                vertices.append(face_vertices)
                faces_dict['bottom'].append(face_vertices)

                # Save the face vertices
                vertex_1 = [self.bottom_surface.x_mesh[row, col], self.bottom_surface.y_mesh[row, col], self.bottom_surface.data[row, col]]
                vertex_2 = [self.bottom_surface.x_mesh[row + 1, col + 1], self.bottom_surface.y_mesh[row + 1, col + 1], self.bottom_surface.data[row + 1, col + 1]]
                vertex_3 = [self.bottom_surface.x_mesh[row + 1, col], self.bottom_surface.y_mesh[row + 1, col], self.bottom_surface.data[row + 1, col]]
                face_vertices = [vertex_1, vertex_2, vertex_3]
                vertices.append(face_vertices)
                faces_dict['bottom'].append(face_vertices)


        # Now, we need to generate the faces for the top surface
        logger.info('Generating faces for the top surface')
        for row in range(self.top_surface.nrows - 1):
            for col in range(self.top_surface.ncols - 1):
                # Generate the face
                # Save the face vertices
                vertex_1 = [self.top_surface.x_mesh[row, col], self.top_surface.y_mesh[row, col], self.top_surface.data[row, col]]
                vertex_2 = [self.top_surface.x_mesh[row, col + 1], self.top_surface.y_mesh[row, col + 1], self.top_surface.data[row, col + 1]]
                vertex_3 = [self.top_surface.x_mesh[row + 1, col + 1], self.top_surface.y_mesh[row + 1, col + 1], self.top_surface.data[row + 1, col + 1]]
                face_vertices = [vertex_1, vertex_2, vertex_3]
                vertices.append(face_vertices)
                faces_dict['top'].append(face_vertices)

                # Save the face vertices
                vertex_1 = [self.top_surface.x_mesh[row, col], self.top_surface.y_mesh[row, col], self.top_surface.data[row, col]]
                vertex_2 = [self.top_surface.x_mesh[row + 1, col + 1], self.top_surface.y_mesh[row + 1, col + 1], self.top_surface.data[row + 1, col + 1]]
                vertex_3 = [self.top_surface.x_mesh[row + 1, col], self.top_surface.y_mesh[row + 1, col], self.top_surface.data[row + 1, col]]
                face_vertices = [vertex_1, vertex_2, vertex_3]
                vertices.append(face_vertices)
                faces_dict['top'].append(face_vertices)

        # Now, we need to generate the four faces that connect the top and bottom surfaces
        # row = 0
        logger.info('Generating faces that connect the top and bottom surfaces')
        for col in range(self.top_surface.ncols - 1):
            # Generate the face
            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[0, col], self.top_surface.y_mesh[0, col], self.top_surface.data[0, col]]
            vertex_2 = [self.top_surface.x_mesh[0, col + 1], self.top_surface.y_mesh[0, col + 1], self.top_surface.data[0, col + 1]]
            vertex_3 = [self.bottom_surface.x_mesh[0, col + 1], self.bottom_surface.y_mesh[0, col + 1], self.bottom_surface.data[0, col + 1]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['north'].append(face_vertices)

            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[0, col], self.top_surface.y_mesh[0, col], self.top_surface.data[0, col]]
            vertex_2 = [self.bottom_surface.x_mesh[0, col + 1], self.bottom_surface.y_mesh[0, col + 1], self.bottom_surface.data[0, col + 1]]
            vertex_3 = [self.bottom_surface.x_mesh[0, col], self.bottom_surface.y_mesh[0, col], self.bottom_surface.data[0, col]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['north'].append(face_vertices)

        # row = nrows - 1
        for col in range(self.top_surface.ncols - 1):
            # Generate the face
            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[-1, col], self.top_surface.y_mesh[-1, col], self.top_surface.data[-1, col]]
            vertex_2 = [self.top_surface.x_mesh[-1, col + 1], self.top_surface.y_mesh[-1, col + 1], self.top_surface.data[-1, col + 1]]
            vertex_3 = [self.bottom_surface.x_mesh[-1, col + 1], self.bottom_surface.y_mesh[-1, col + 1], self.bottom_surface.data[-1, col + 1]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['south'].append(face_vertices)

            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[-1, col], self.top_surface.y_mesh[-1, col], self.top_surface.data[-1, col]]
            vertex_2 = [self.bottom_surface.x_mesh[-1, col + 1], self.bottom_surface.y_mesh[-1, col + 1], self.bottom_surface.data[-1, col + 1]]
            vertex_3 = [self.bottom_surface.x_mesh[-1, col], self.bottom_surface.y_mesh[-1, col], self.bottom_surface.data[-1, col]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['south'].append(face_vertices)

        # col = 0
        for row in range(self.top_surface.nrows - 1):
            # Generate the face
            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[row, 0], self.top_surface.y_mesh[row, 0], self.top_surface.data[row, 0]]
            vertex_2 = [self.top_surface.x_mesh[row + 1, 0], self.top_surface.y_mesh[row + 1, 0], self.top_surface.data[row + 1, 0]]
            vertex_3 = [self.bottom_surface.x_mesh[row + 1, 0], self.bottom_surface.y_mesh[row + 1, 0], self.bottom_surface.data[row + 1, 0]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['west'].append(face_vertices)

            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[row, 0], self.top_surface.y_mesh[row, 0], self.top_surface.data[row, 0]]
            vertex_2 = [self.bottom_surface.x_mesh[row + 1, 0], self.bottom_surface.y_mesh[row + 1, 0], self.bottom_surface.data[row + 1, 0]]
            vertex_3 = [self.bottom_surface.x_mesh[row, 0], self.bottom_surface.y_mesh[row, 0], self.bottom_surface.data[row, 0]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['west'].append(face_vertices)

        # col = ncols - 1
        for row in range(self.top_surface.nrows - 1):
            # Generate the face
            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[row, -1], self.top_surface.y_mesh[row, -1], self.top_surface.data[row, -1]]
            vertex_2 = [self.top_surface.x_mesh[row + 1, -1], self.top_surface.y_mesh[row + 1, -1], self.top_surface.data[row + 1, -1]]
            vertex_3 = [self.bottom_surface.x_mesh[row + 1, -1], self.bottom_surface.y_mesh[row + 1, -1], self.bottom_surface.data[row + 1, -1]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['east'].append(face_vertices)

            # Save the face vertices
            vertex_1 = [self.top_surface.x_mesh[row, -1], self.top_surface.y_mesh[row, -1], self.top_surface.data[row, -1]]
            vertex_2 = [self.bottom_surface.x_mesh[row + 1, -1], self.bottom_surface.y_mesh[row + 1, -1], self.bottom_surface.data[row + 1, -1]]
            vertex_3 = [self.bottom_surface.x_mesh[row, -1], self.bottom_surface.y_mesh[row, -1], self.bottom_surface.data[row, -1]]
            face_vertices = [vertex_1, vertex_2, vertex_3]
            vertices.append(face_vertices)
            faces_dict['east'].append(face_vertices)

        vertices = np.array(vertices)

        # Now we need to generate the mesh
        self.stl_file = mesh.Mesh(np.zeros(vertices.shape[0], dtype=mesh.Mesh.dtype))
        for i, f in enumerate(vertices):
            for j in range(3):
                self.stl_file.vectors[i][j] = f[j]

        # Create STL meshes for the faces
        self.faces_stl = {}
        for face in faces_dict:
            vertices = np.array(faces_dict[face])
            self.faces_stl[face] = mesh.Mesh(np.zeros(vertices.shape[0], dtype=mesh.Mesh.dtype))
            for i, f in enumerate(vertices):
                for j in range(3):
                    self.faces_stl[face].vectors[i][j] = f[j]

        self.export_stl(output_filename=output_filename, faces=export_faces)



    def export_stl(self, output_filename, faces=False):
        """
        Exports the generated closed STL
        """
        from pathlib import Path
        results_folder = create_results_folder()
        output_filename = Path(output_filename)
        logger.info(f'Writing mesh to {output_filename}')
        self.stl_file.save(str(results_folder / output_filename))
        if faces:
            for face in self.faces_stl:
                face_filename = Path(f"{output_filename.stem}-{face}.stl")
                logger.info(f'Writing {face} face STL mesh to outputs folder')
                self.faces_stl[face].save(str(results_folder / face_filename))





    def plot_aperture(self):
        """This method plots the aperture"""
        self.aperture.plot(colorbar_label='Aperture')




from pydelling.managers import BaseManager, BaseStudy
from abc import ABC, abstractmethod
import logging

logger = logging.getLogger(__name__)


class BaseCallback(ABC):
    """Base class for callbacks."""
    def __init__(self, manager: BaseManager, study: BaseStudy, kind: str = None, **kwargs):
        self.manager = manager
        self.study = study
        assert kind in ['pre', 'post'], f"Kind must be 'pre' or 'post', not {kind}"
        self.kind = kind
        self.kwargs = kwargs
        self.is_run = False
        self.process_kwargs()

    @abstractmethod
    def run(self):
        """This method executes the callback."""
        self.is_run = True
        logger.info(f"Running callback {self.__class__.__name__} for study {self.study.name}")
        pass

    def process_kwargs(self):
        for kwarg_name, kwarg in self.kwargs.items():
            setattr(self, kwarg_name, kwarg)






import unittest
import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Line
from pydelling.preprocessing import DfnPreprocessor, MeshPreprocessor
from pydelling.preprocessing.dfn_preprocessor import DfnUpscaler


class UpscalingCase(unittest.TestCase):
    def setUp(self) -> None:
        self.dfn_preprocessor = DfnPreprocessor()
        self.dfn_preprocessor.add_fracture(
            x=0.0, y=0.0, z=0.0, dip=0, dip_dir=0, size=2.0
        )
        self.dfn_preprocessor.add_fracture(
            x=0.1, y=0.0, z=0.1, dip=0, dip_dir=0, size=0.5
        )
        self.dfn_preprocessor[0].aperture = 0.01
        self.dfn_preprocessor[1].aperture = 0.01

        self.dfn_preprocessor[0].hydraulic_aperture = 0.01
        self.dfn_preprocessor[1].hydraulic_aperture = 0.01

        self.dfn_preprocessor.to_obj('fracture.obj')
        self.mesh_preprocessor = MeshPreprocessor()
        self.mesh_preprocessor.add_hexahedra(
            node_ids=np.array([0, 1, 2, 3, 4, 5, 6, 7]),
            node_coords=[
                np.array([-0.5, -0.5, -0.5]),
                np.array([0.5, -0.5, -0.5]),
                np.array([0.5, 0.5, -0.5]),
                np.array([-0.5, 0.5, -0.5]),
                np.array([-0.5, -0.5, 0.5]),
                np.array([0.5, -0.5, 0.5]),
                np.array([0.5, 0.5, 0.5]),
                np.array([-0.5, 0.5, 0.5])
            ],
        )
        self.mesh_preprocessor.to_vtk(filename='./test_mesh.vtk')

        self.dfn_upscaler = DfnUpscaler(
            dfn=self.dfn_preprocessor,
            mesh=self.mesh_preprocessor,
        )

    def test_porosity_and_permeability(self):
        porosity = self.dfn_upscaler.upscale_mesh_porosity(intensity_correction_factor=1.53,
                                                           existing_fractures_fraction=0.385,
                                                           truncate_to_min_percentile=0,
                                                           truncate_to_max_percentile=100)

        porosity_solution = 0.049675324675324685
        self.assertAlmostEqual(porosity[0], porosity_solution)
        # Upscale permeability
        permeability = self.dfn_upscaler.upscale_mesh_permeability(truncate_to_min_percentile=0,
                                                                   truncate_to_max_percentile=100)
        # #self.assertEqual(porosity, porosity_solution)
        permeability_solution = np.array([[1.1470037453,0,0],[0,1.1470037453,0],[0,0,0]])
        nptest.assert_array_almost_equal(permeability[0], permeability_solution)

    # def test_save_load_upscaler(self):
    #     self.dfn_upscaler.save('upscaler.pkl')
    #     new_upscaler = DfnUpscaler.load('upscaler.pkl')
    #     self.assertEqual(len(self.dfn_upscaler.mesh.elements), len(new_upscaler.mesh.elements))
    #     self.assertEqual(len(self.dfn_upscaler.dfn.dfn), len(new_upscaler.dfn.dfn))
    #     self.assertEqual(len(self.dfn_upscaler.dfn.dfn[0].intersection_dictionary), len(new_upscaler.dfn.dfn[0].intersection_dictionary))
    #     self.assertEqual(len(self.dfn_upscaler.dfn.faults), len(new_upscaler.dfn.faults))
    #     self.assertEqual(len(self.dfn_upscaler.upscaled_porosity), len(new_upscaler.upscaled_porosity))
    #     from pathlib import Path
    #     Path('upscaler.pkl').unlink()

    # def test_save_load_json(self):
    #     self.dfn_upscaler.to_json('upscaler.json')
    #     new_upscaler = DfnUpscaler.from_json('upscaler.json')
    #     new_upscaler.upscale_mesh_porosity()
    #     # self.dfn_upscaler.upscale_mesh_porosity()
    #     # self.assertEqual(self.dfn_upscaler.upscaled_porosity[0], new_upscaler.upscaled_porosity[0])
    # #


if __name__ == '__main__':
    unittest.main()


from abc import ABC, abstractmethod
import pandas as pd
import paramiko
import logging
logger = logging.getLogger(__name__)
from pathlib import Path

class BaseSsh(ABC):
    client: paramiko.SSHClient
    sftp: paramiko.SFTPClient
    def __init__(self, user,
                 pkey_path,
                 password=None):
        """
        Connects to the remote server.
        Args:
            user: username
            pkey_path: path to the private key
        """
        self.pkey_path = pkey_path
        self.user = user
        self.password = password
        self.connect()

    @abstractmethod
    def connect(self):
        pass

    def run_command(self, command):
        """
        Runs a ssh command in the remote server.
        Args:
            command: command to run

        Returns: stdout

        """
        stdin, stdout, stderr = self.client.exec_command(command)
        return stdout.read().decode('utf-8').strip()

    @property
    def pwd(self):
        """
        Returns the current working directory in the remote server.
        Returns: current working directory
        """
        return self.sftp.getcwd()

    def cd(self, path):
        """
        Changes the current working directory in the remote server.
        Args:
            path: path to change to
            """
        path = str(path)
        self.sftp.chdir(path)
        logger.info(f'Changed directory to {path}')

    def mkdir(self, path):
        """
        Creates a directory in the remote server.
        Args:
            path: path to create
        """
        path = Path(path)
        if path.name in self.ls:
            logger.info(f'Directory {path} already exists')
            return
        else:
            self.sftp.mkdir(str(path))
            logger.info(f'Created directory {path}')

    def rm(self, path):
        """
        Removes a file in the remote server.
        Args:
            path: file to remove

        Returns:

        """
        path = str(path)
        self.sftp.remove(path)
        logger.info(f'Removed file {path}')

    def rmdir(self,
              path,
              ):
        """
        Removes a directory in the remote server.

        Args:
            path: directory to remove

        Returns:

        """
        path = Path(path)
        for file in self.sftp.listdir(str(path)):
            try:
                self.rm(str(path / file))
            except IOError:
                self.rmdir(str(path / file))
        self.sftp.rmdir(str(path))
        logger.info(f'Removed directory {path}')

    def cp(self, src, dst):
        """
        Copies a file from the local machine to the remote server.
        Args:
            src: source file
            dst: destination file
        """
        self.sftp.put(src, dst)
        logger.info(f'Copied {src} to {dst}')

    def cpdir(self, src, dst):
        """
        Copies a directory from the local machine to the remote server.
        Args:
            src: source directory
            dst: destination directory
        """
        src = Path(src)
        dst = Path(dst)

        self.mkdir(str(dst))
        for file in src.iterdir():
            if file.is_dir():
                self.cpdir(file, str(dst / file.name))
            else:
                self.cp(file, str(dst / file.name))

    def get(self, src, dst):
        """
        Copies a file from the remote server to the local machine.
        Args:
            src: source file
            dst: destination file
        """
        self.sftp.get(src, dst)
        logger.info(f'Copied {src} to {dst}')

    def getdir(self, src, dst):
        """
        Copies a directory from the remote server to the local machine.
        Args:
            src: source directory
            dst: destination directory
        """
        src = Path(src)
        dst = Path(dst)

        dst.mkdir(parents=True, exist_ok=True)
        self.cd(src)
        for file in self.ls:
            self.get(file, dst / Path(file).name)


    @property
    def ls(self):
        """
        Returns the content of the current working directory in the remote server.
        Returns: content of the current working directory
        """
        return self.sftp.listdir()




from pydelling.webapps.components import BaseComponent
import streamlit as st
import time
import paramiko
import extra_streamlit_components as stx


@st.cache(allow_output_mutation=True)
def get_manager():
    return stx.CookieManager()

manager = get_manager()

class RemoteLoginComponent(BaseComponent):
    def __int__(self,
                host: str,
                host_name: str = None,
                username: str = None,
                password: str = None,
                login_node: bool = False,
                *args,
                **kwargs
                ):
        self.initialize_in_session_state('host')
        self.initialize_in_session_state('host_name')
        self.initialize_in_session_state('username')
        self.initialize_in_session_state('cwd')
        self.initialize_in_session_state('password')
        self.host = host
        self.host_name = host_name if host_name is not None else host
        self.client: paramiko.SSHClient
        self.username = username
        self.password = password
        self.login_node = login_node
        self.cookie_manager = self.get_manager()
        super().__init__(host=host,
                         host_name=host_name,
                         username=username,
                         password=password,
                         login_node=login_node,
                         *args,
                         **kwargs,
                         )

    def run(self, login_node=True, *args, **kwargs):
        if not self.get_from_session_state(f'{self.name}-init') and login_node:
            with st.form(key='login_form'):
                st.markdown(f'Login form for {self.host_name} ({self.host})')
                col1, col2 = st.columns(2)
                with col1:
                    self.username = st.text_input('Username')
                    cwd = st.text_input('Working directory')
                    submitted = st.form_submit_button('Log in')
                with col2:
                    self.password = st.text_input('Password', type='password')
                if submitted:
                    self.login_submit_func(username=self.username, cwd=cwd, password=self.password)
                    st.experimental_rerun()
        else:
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            # Just run the component logic

    # @st.experimental_memo
    def login_submit_func(self, username, cwd, password):
        # Log in to system and be sure it works
        # self = _self
        with st.spinner('Logging in...'):
            self.cookie_manager: stx.CookieManager = self.get_manager()
            self.save_in_session_state(f'host', self.host)
            self.save_in_session_state(f'host_name', self.host_name)
            self.save_in_session_state(f'username', username)
            self.save_in_session_state(f'password', self.password)
            self.save_in_session_state(f'cwd', cwd)

            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.client.connect(self.host, username=username, password=password)
            st.success(f'You are logged in to {self.host}')
            auth_dict = {
                'host': self.host,
                'host_name': self.host_name,
                'username': username,
                'password': password,
                'cwd': cwd,
            }
            get_manager().set(f'{self.host_name}-auth', auth_dict)
            time.sleep(0.3)
            self.save_in_session_state('is_login', True)
            self.save_in_session_state(f'{self.name}-init', True)

    def check_connection(self):
        # Check if the connection is still alive
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.connect(self.host,
                            username=self.username,
                            password=self.password,
                            )
        return True






import unittest
from pydelling.readers.FemReader import FemReader
from pydelling.utils import test_data_path


class TestFemReader(unittest.TestCase):
    def test_read_data(self):
        fem_reader = FemReader(test_data_path() / "fem_reader_data.fem")
        self.assertEqual(fem_reader.n_nodes, 1373)
        self.assertEqual(fem_reader.n_elements, 3909)


if __name__ == '__main__':
    unittest.main()


import os

class BaseStatus:
    """The base class for all status classes. This class should read a given status file and extract some key information."""

    def __init__(self, status_file, ssh_manager=None):
        """Initialize the BaseStatus class with the given status file."""
        if not os.path.exists(status_file):
            raise FileNotFoundError(f"Status file {status_file} not found.")
        self.status_file = status_file
        self.status_data = None
        self.ssh_manager = ssh_manager
        self.read_status_file()

    def read_status_file(self):
        """Read the status file and store its content."""
        # if self.ssh_manager is not None:
        #     self.status_data = self.ssh_manager.get(self.status_file
        with open(self.status_file, "r") as file:
            self.status_data = file.read()

    def extract_key_information(self):
        """Extract key information from the status data."""
        raise NotImplementedError("This method should be implemented in a derived class.")

    @property
    def progress(self):
        """Return the progress of the simulation."""
        raise NotImplementedError("This method should be implemented in a derived class.")



from .BasePreprocessing import BasePreprocessing
# from .STLFromPointCloud import STLFromPointCloud
from .dfn_preprocessor import DfnPreprocessor
from .mesh_preprocessor import MeshPreprocessor
from .ClosedStlGenerator import ClosedStlGenerator




"""
Base interface for a reader class
"""
import logging

import numpy as np

from pydelling.readers import PflotranObservationPointReader

logger = logging.getLogger(__name__)
from pydelling.config import config
import logging
from pathlib import Path
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
logger = logging.getLogger(__name__)


class PflotranMassBalanceFileReader(PflotranObservationPointReader):
    observation_point: np.ndarray
    observation_boundary: str
    observation_node: int
    variables: dict

    def __init__(self, filename=None):
        self.filename = Path(filename) if filename else Path(config.pflotran_reader.filename)
        logger.info(f"Reading PFLOTRAN mass balance file from {self.filename}")
        super().__init__(filename=self.filename)
        self.results = self.data.copy()

    def generate_mass_balance_plot(self) -> plt.Axes:
        """
        Generate a mass balance plot

        Returns
        -------
        plt.Axes
            The mass balance plot
        """




import logging

import numpy as np
from typing import Tuple, List, Union

from .BaseReader import BaseReader

logger = logging.getLogger(__name__)


class RasterFileReader(BaseReader):
    """
    Class that contains functions to read a rasterized file in .asc format
    """
    def __init__(self,
                 filename=None,
                 header=False,
                 read_data=True,
                 **kwargs,):
        super().__init__(filename=filename, header=header, read_data=read_data, **kwargs)

    def open_file(self, filename, n_header=6):
        with open(filename, 'r') as opened_file:
            if self.header:
                opened_file.readline()  # For now, skips the header if it has
            self.read_file(opened_file, n_header)
        self.build_info()

    def read_file(self, opened_file, n_header=6):
        self.read_header(opened_file, n_header=n_header)
        self.build_structure()
        self.read_data(opened_file)
        logger.info(f"Reading ASC raster file from {opened_file}")

    def read_header(self, opened_file, n_header=6):
        flag_stop = False
        while not flag_stop:
            current_position = opened_file.tell()
            line = opened_file.readline().split()
            if len(line) != 2:
                opened_file.seek(current_position)
                flag_stop = True

            self.info["reader"][line[0]] = float(line[1])
        if 'cellsize' in self.info["reader"].keys():
            self.info["reader"]["dx"] = self.info["reader"]["cellsize"]
            self.info["reader"]["dy"] = self.info["reader"]["cellsize"]


    def read_data(self, opened_file):
        for id, line in enumerate(opened_file.readlines()):
            self.data[id] = np.array(line.split(), dtype=np.float32)

    def build_structure(self):
        assert self.info is not {}
        self.xydata_computed = False
        self.data = np.zeros(shape=(int(self.reader_info["nrows"]), int(self.reader_info["ncols"])))

        if 'cellsize' in self.reader_info:
            x_range = np.arange(self.reader_info["xllcorner"], self.reader_info["xllcorner"] + self.reader_info["nrows"] * self.reader_info["cellsize"],
                                self.reader_info["cellsize"])
            y_range = np.arange(self.reader_info["yllcorner"], self.reader_info["yllcorner"] + self.reader_info["ncols"] * self.reader_info["cellsize"],
                                self.reader_info["cellsize"])
            print(self.reader_info['xllcorner'], self.reader_info['yllcorner'], self.reader_info['cellsize'], self.reader_info['nrows'], self.reader_info['ncols'])
        else:
            x_range = np.arange(self.reader_info["xllcorner"], self.reader_info["xllcorner"] + self.reader_info["nrows"] * self.reader_info["dx"],
                                self.reader_info["dx"])
            y_range = np.arange(self.reader_info["yllcorner"], self.reader_info["yllcorner"] + self.reader_info["ncols"] * self.reader_info["dy"],
                                self.reader_info["dy"])
        self.x_mesh, self.y_mesh = np.meshgrid(x_range, y_range)
        self.y_mesh = np.flipud(self.y_mesh)  # To fit into the .asc format criteria

    def build_info(self):
        """
        Function that creates the internal data structure of the raster file
        """
        self.info["reader"]["filename"] = self.filename

    def add_z_info(self, z_coord):
        self.z_coord = z_coord

    def get_data(self) -> np.ndarray:
        # self.rebuild_x_y()
        if hasattr(self, "z_coord"):
            return self.get_xyz_data()
        else:
            return self.get_xy_data()

    def rebuild_x_y(self):
        if 'cellsize' in self.info['reader']:
            x_range = np.arange(self.info['reader']["xllcorner"],
                                self.info['reader']["xllcorner"] + self.info['reader']["nrows"] * self.info['reader']["cellsize"],
                                self.info['reader']["cellsize"])
            y_range = np.arange(self.info['reader']["yllcorner"],
                                self.info['reader']["yllcorner"] + self.info['reader']["ncols"] * self.info['reader']["cellsize"],
                                self.info['reader']["cellsize"])
        else:
            x_range = np.arange(self.info['reader']["xllcorner"],
                                self.info['reader']["xllcorner"] + self.info['reader']["nrows"] * self.info['reader']["dx"],
                                self.info['reader']["dx"])
            y_range = np.arange(self.info['reader']["yllcorner"],
                                self.info['reader']["yllcorner"] + self.info['reader']["ncols"] * self.info['reader']["dy"],
                                self.info['reader']["dy"])
        self.x_mesh, self.y_mesh = np.meshgrid(x_range, y_range)
        self.y_mesh = np.flipud(self.y_mesh)  # To fit into the .asc format criteria


    def get_xy_data(self) -> np.ndarray:
        ndata = int(self.info["reader"]["nrows"] * self.info["reader"]["ncols"])
        self.xydata = np.zeros(shape=(ndata, 3))
        rows_mesh_flatten = np.fliplr(self.x_mesh).T.flatten()
        cols_mesh_flatten = np.flipud(self.y_mesh).T.flatten()
        for id, data in enumerate(self.data.flatten()):
            self.xydata[id] = (cols_mesh_flatten[id], rows_mesh_flatten[id], data)
        self.xydata_computed = True
        return self.xydata

    def get_xyz_data(self) -> np.ndarray:
        assert hasattr(self, "z_coord"), "The z-coordinate of this raster file is not given"
        ndata = int(self.info["reader"]["nrows"] * self.info["reader"]["ncols"])
        self.flatten_data = np.zeros(shape=(ndata, 4))
        rows_mesh_flatten = np.fliplr(self.x_mesh).T.flatten()
        cols_mesh_flatten = np.flipud(self.y_mesh).T.flatten()
        for id, data in enumerate(self.data.flatten()):
            self.flatten_data[id] = (cols_mesh_flatten[id], rows_mesh_flatten[id], self.z_coord, data)
        self.xydata_computed = True
        return self.flatten_data

    def to_csv(self, output_file, z_coord=None):
        """
        Function that writes the ratser data into a csv file
        :param output_file:
        :return:
        """

        xydata = self.get_xy_data()
        f = open(output_file, "w")
        if z_coord is not None:
            f.write(f"x,y,z,data\n")
        else:
            f.write(f"x,y,data\n")
        for data in xydata:
            if z_coord is not None:
                f.write(f"{data[0]},{data[1]},{z_coord},{data[2]}\n")
            else:
                f.write(f"{data[0]},{data[1]},{data[2]}\n")
        f.close()
        logger.info(f"The raster file points have been exported to the CSV file {output_file}")

    def to_wsv(self, output_file):
        print(f"Starting dump into {output_file}")
        if not self.xydata_computed:
            xydata = self.dump_to_xydata()
        else:
            xydata = self.xydata
        f = open(output_file, "w")
        for data in xydata:
            f.write(f"{data[0]} {data[1]} {data[2]}\n")
        f.close()
        print(f"The data has been properly exported to the {output_file} file")

    def to_asc(self, output_file):
        logger.info(f"Starting dump into {output_file}")
        with open(output_file, 'w') as file:
            self.write_asc_header(file)
            self.write_asc_data(file)

    def write_asc_header(self, file):
        # assert isinstance(file, type(open)), "is not a correct file"
        # Write info in ASC format
        file.write(f"ncols {self.info['reader']['ncols']}\n")
        file.write(f"nrows {self.info['reader']['nrows']}\n")
        if 'cellsize' in self.info:
            file.write(f"cellsize {self.info['reader']['cellsize']}\n")
        else:
            file.write(f"dx {self.info['reader']['dx']}\n")
            file.write(f"dy {self.info['reader']['dy']}\n")
        file.write(f"xllcorner {self.info['reader']['xllcorner']}\n")
        file.write(f"yllcorner {self.info['reader']['yllcorner']}\n")
        file.write(f"NODATA_value {self.info['reader']['NODATA_value']}\n")


    def write_asc_data(self, file):
        np.savetxt(file, self.data)

    def downsample_data(self, slice_factor=2):
        '''
        This module downsamples the data based on a constant stride in each direction
        :param slice_factor: factor to stride the matrix in each dimension

        :return: downsampled dataset
        '''
        self.data = self.data[0::slice_factor, 0::slice_factor]
        self.info['reader']["nrows"] = self.data.shape[0]
        self.info['reader']["ncols"] = self.data.shape[1]
        if "cellsize" in self.info["reader"]:
            self.info['reader']["cellsize"] *= slice_factor
        else:
            self.info['reader']["dx"] *= slice_factor
            self.info['reader']["dy"] *= slice_factor
        print(self.data.shape)
        self.rebuild_x_y()
        logger.info(f"Data has been downsampled by a factor of {slice_factor}")

    def get_value_from_coord(self, x: float, y: float):
        """
        This method returns the raster value finding the nearest neighbour of a given x, y coordinate
        Args:
            x: Coordinate of the x-direction
            y: Coordinate of the y-direction

        Returns:
            Raster value at the given coordinate
        """
        # Define auxiliaty variables
        d_raster = self.reader_info["cellsize"]
        origin_x = self.reader_info["xllcorner"]
        origin_y = self.reader_info["yllcorner"]

        ix = int(np.floor((x - origin_x) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
        # the floor function
        iy = int(np.floor((y - origin_y) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
        # the floor function
        iy = int(self.reader_info["ncols"] - iy - 1)
        return self.data[iy, ix]

    def flip_y(self):
        self.info['reader']["yllcorner"] = self.info['reader']["yllcorner"] + self.info['reader']["cellsize"] * self.info['reader']["nrows"]
        RasterFileReader.rebuild_x_y(self)

    def flip_x(self):
        self.info['reader']["xllcorner"] = self.info['reader']["xllcorner"] + self.info['reader']["cellsize"] * self.info['reader']["ncols"]
        RasterFileReader.rebuild_x_y(self)

    def get_plot_image(self, ax=None,
                       fig=None,
                       colorbar=True,
                       colorbar_label=None,
                       **kwargs):
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots()
        ax.imshow(self.data, **kwargs)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        if colorbar:
            fig.colorbar(ax.get_images()[0], label=colorbar_label)
        return ax

    def plot(self, colorbar=True, colorbar_label=None, **kwargs):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        self.get_plot_image(ax=ax,
                            fig=fig,
                            colorbar=colorbar,
                            colorbar_label=colorbar_label,
                            **kwargs)
        plt.show()

    def save_plot(self, output_file, colorbar=None, colorbar_label=None,**kwargs):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        self.get_plot_image(ax=ax,
                            fig=fig,
                            colorbar=colorbar,
                            colorbar_label=colorbar_label,
                            **kwargs)
        plt.savefig(output_file)

    @property
    def nx(self) -> int:
        return int(self.info['reader']["ncols"])

    @property
    def ny(self) -> int:
        return int(self.info['reader']["nrows"])

    @property
    def nrows(self) -> int:
        return int(self.info['reader']["nrows"])

    @property
    def ncols(self) -> int:
        return int(self.info['reader']["ncols"])

    @property
    def dx(self):
        if 'dx' in self.info['reader']:
            return self.info['reader']['dx']
        else:
            return self.info['reader']["cellsize"]

    @property
    def dy(self):
        if 'dy' in self.info['reader']:
            return self.info['reader']['dy']
        else:
            return self.info['reader']["cellsize"]

    # Read raster file from a .csv file
    @classmethod
    def from_xyz_csv(self, file_name, dims=None,):
        data = np.loadtxt(file_name, delimiter=",", skiprows=1)
        if dims is None:
            dims = self._guess_raster_dimensions_from_data(data)
        else:
            assert dims[0] * dims[1] == data.shape[0], "The dimensions are not correct"
            dims = (dims[1], dims[0])

        info = {'reader': {}}
        info['reader']["ncols"] = dims[1]
        info['reader']["nrows"] = dims[0]
        info['reader']["xllcorner"] = data[:, 0].min()
        info['reader']["yllcorner"] = data[:, 1].min()
        info['reader']['dx'] = data[1, 0] - data[0, 0]
        info['reader']['dy'] = data[dims[1], 1] - data[0, 1]
        info['reader']['dx'] = info['reader']['dx']
        info['reader']['dy'] = - info['reader']['dy']
        info['reader']["NODATA_value"] = -9999
        return_raster = RasterFileReader(filename=file_name, data=data, info=info, read_data=False)
        return_raster.build_structure()
        return_raster.data = data[:, 2].reshape(dims[0], dims[1])
        return return_raster

    @staticmethod
    def _guess_raster_dimensions_from_data(data: np.ndarray) -> Tuple[int, int]:
        """
        This method guesses the raster dimensions from the data
        Args:
            data: array with the raster data

        Returns:
            Tuple with the number of rows and columns
        """
        if data.shape[0] == data.shape[1]:
            # Square raster
            return data.shape[0], data.shape[1]
        else:
            # Assume dimensions are ordered in x, y, target
            first_x_value = data[0, 0]
            first_y_value = data[0, 1]
            steps_with_same_x = 1
            steps_with_same_y = 1

            for idx in range(1, data.shape[0]):
                if data[idx, 0] == first_x_value:
                    steps_with_same_x += 1
                if data[idx, 1] == first_y_value:
                    steps_with_same_y += 1
            assert steps_with_same_x * steps_with_same_y == data.shape[0], f"Cant guess dimensions from data. Computed dimensions are {steps_with_same_x}x{steps_with_same_y}={steps_with_same_x * steps_with_same_y} and data has {data.shape[0]} elements"
            return steps_with_same_x, steps_with_same_y

    def get_data_from_coordinates(self, x, y):
        """
        This method returns the data from the raster at the given coordinates
        Args:
            x: x coordinate
            y: y coordinate

        Returns:
            The value of the raster at the given coordinates
        """
        x_idx = np.argmin(np.abs(self.y - y))
        y_idx = np.argmin(np.abs(self.x - x))
        return self.data[self.nrows - x_idx - 1, y_idx]

    @property
    def x(self):
        return self.info['reader']["xllcorner"] + np.arange(self.nx) * self.dx

    @property
    def y(self):
        return self.info['reader']["yllcorner"] + np.arange(self.ny) * self.dy

    # Add difference of two raster files
    def __sub__(self, other):
        if isinstance(other, RasterFileReader):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data - other.data)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]
            return new_raster
        elif isinstance(other, (int, float)):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data - other)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]

            return new_raster
        else:
            raise ValueError("The other object is not a RasterFileReader or a number")

    # Add sum of two raster files
    def __add__(self, other):
        if isinstance(other, RasterFileReader):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data + other.data)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]

            return new_raster
        elif isinstance(other, (int, float)):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data + other)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]

            return new_raster
        else:
            raise ValueError("The other object is not a RasterFileReader or a number")

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data * other)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]
            return new_raster

        elif isinstance(other, RasterFileReader):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data * other.data)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]

            return new_raster

        else:
            raise ValueError("The other object is not a RasterFileReader or a number")

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data / other)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]
            return new_raster

        elif isinstance(other, RasterFileReader):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data / other.data)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]

            return new_raster

        else:
            raise ValueError("The other object is not a RasterFileReader or a number")

    def __pow__(self, other):
        if isinstance(other, (int, float)):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data ** other)
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]
            return new_raster

        elif isinstance(other, RasterFileReader):
            new_raster = RasterFileReader(filename=self.filename, read_data=False, info=self.info, data=self.data ** other.data)
            new_raster.info = self.info
            for key in self.__dict__.keys():
                if key != "data" and key != "info":
                    new_raster.__dict__[key] = self.__dict__[key]


            return new_raster

        else:
            raise ValueError("The other object is not a RasterFileReader or a number")

    @property
    def reader_info(self):
        return self.info['reader']













import unittest
from pydelling.managers.status import PflotranStatus
from pydelling.utils import test_data_path


class TestPflotranStatus(unittest.TestCase):
    def setUp(self) -> None:
        self.pflotran_status = PflotranStatus(test_data_path() / 'pflotran_status.out')

    def test_class(self):
        self.assertEqual(self.pflotran_status.wall_clock_time, 27361.0)
        self.assertTrue(self.pflotran_status.is_done)
        self.assertEqual(len(self.pflotran_status.times), len(self.pflotran_status.dts))
        self.assertIsNone(self.pflotran_status.progress)
        self.pflotran_status.add_total_time(11000.0)
        self.assertEqual(self.pflotran_status.progress, 1.0)


if __name__ == '__main__':
    unittest.main()





from __future__ import annotations

import logging
from typing import List

import numpy as np

from .Point import Point
from ._AbstractGidObject import _AbstractGidObject

logger = logging.getLogger(__name__)

class Line(_AbstractGidObject):
    local_id: int = 1
    global_id: int = 1
    lines: List = []
    has_copy = False
    def __init__(self, point_1: Point, point_2: Point):
        self.local_id = Line.local_id
        self.id = None
        if np.array_equal(point_1.coords, point_2.coords):
            raise ValueError('The start and end points of a line cannot be equal')
        self.point_1 = point_1
        self.point_2 = point_2
        self.points = [self.point_1, self.point_2]
        Line.local_id += 1

    def add(self):
        # Check if line already exists
        for line in Line.lines:
            if Line.check_lines_equal(self, line):
                self.id = line.id
                self.has_copy = True
                return ''

        self.id = Line.global_id
        Line.global_id += 1
        Line.lines.append(self)
        logger.info(f'Adding line {self.id} connecting {self.point_1} and {self.point_2}')
        export_str = 'Mescape Geometry Create Line\n'
        export_str += 'Join\n'
        export_str += f'{self.point_1.id}\n'
        export_str += f'{self.point_2.id}\n'
        return export_str

    @staticmethod
    def check_lines_equal(line_1: Line, line_2: Line):
        point_1_coords = [line_1.point_1.coords.tolist(), line_1.point_2.coords.tolist()]
        if line_2.point_1.coords.tolist() in point_1_coords and line_2.point_2.coords.tolist() in point_1_coords:
            return True
        else:
            return False

    def __repr__(self):
        if self.id:
            return f"Line {self.id}"
        else:
            return f"Line (local) {self.local_id}"










import numpy as np

from pydelling.readers.iGPReader.geometry import BaseFace


class TriangleFace(BaseFace):
    def __init__(self, nodes, coords):
        super().__init__(nodes, coords)
        self.type = "Triangle"

    def compute_centroid(self):
        return np.mean(self.coords, axis=0)



import unittest
from pydelling.readers import RasterFileReader
from pydelling.utils import test_data_path
import numpy.testing as npt


class TestSMeshMeshReader(unittest.TestCase):
    def setUp(self) -> None:
        self.top_surface = RasterFileReader(test_data_path() / 'top_surface.asc')
        self.bottom_surface = RasterFileReader(test_data_path() / 'bottom_surface.asc')
        self.bottom_surface_irregular = RasterFileReader(test_data_path() / 'bottom_surface_irregular.asc')
    def test_read_data(self):
        self.assertEqual(self.top_surface.nx, 250)
        self.assertEqual(self.top_surface.ny, 250)

    def test_write_read_irregular(self):
        self.bottom_surface_irregular.to_asc('test.asc')
        import matplotlib.pyplot as plt
        new_raster = RasterFileReader('test.asc')
        self.assertEqual(self.bottom_surface_irregular.nx, new_raster.nx)
        self.assertEqual(self.bottom_surface_irregular.ny, new_raster.ny)
        # Check the data is the same
        npt.assert_array_equal(self.bottom_surface_irregular.data, new_raster.data)




    def test_operations(self):
        # constant sum
        sum = self.top_surface + 1
        self.assertEqual(self.top_surface.values[0, 2] + 1, sum.values[0, 2])
        # Constant subtraction
        diff = self.top_surface - 1
        self.assertEqual(self.top_surface.values[:, 2].mean() - 1, diff.values[:, 2].mean())
        # Constant multiplication
        prod = self.top_surface * 2
        self.assertEqual(self.top_surface.values[:, 2].mean() * 2, prod.values[:, 2].mean())
        # Constant division
        quot = self.top_surface / 2
        self.assertEqual(self.top_surface.values[:, 2].mean() / 2, quot.values[:, 2].mean())
        # Raster sum
        sum = self.top_surface + self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] + self.bottom_surface.values[0, 2], sum.values[0, 2])
        # Raster subtraction
        diff = self.top_surface - self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] - self.bottom_surface.values[0, 2], diff.values[0, 2])
        # Raster multiplication
        prod = self.top_surface * self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] * self.bottom_surface.values[0, 2], prod.values[0, 2])
        # Raster division
        quot = self.top_surface / self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] / self.bottom_surface.values[0, 2], quot.values[0, 2])

    def test_get_data_from_coordinates(self):
        data = self.top_surface.get_data_from_coordinates(0.001240760000, 0.199278)
        self.assertAlmostEqual(data, 0.1182)

    def test_downsample(self):
        self.top_surface.downsample_data(2)
        self.assertEqual(self.top_surface.nx, 125)
        self.assertEqual(self.top_surface.ny, 125)

    def tearDown(self) -> None:
        from pathlib import Path
        if Path('test.asc').exists():
            Path('test.asc').unlink()



if __name__ == '__main__':
    unittest.main()


# import logging
# from pathlib import Path
#
# import numpy as np
# import open3d as o3d
# import pandas as pd
# import vtk
#
# from pydelling.config import config
# from pydelling.preprocessing import BasePreprocessing
# from pydelling.utils.decorators import set_run
#
# logger = logging.getLogger(__name__)
#
#
# class STLFromPointCloud(BasePreprocessing):
#     preprocessed_data: np.ndarray
#     is_run: bool = False
#     stl_mesh: o3d.geometry.TriangleMesh
#
#     def __init__(self, data: pd.DataFrame = None, filename: str = None):
#         super().__init__(data, filename)
#
#
#     @set_run
#     def run(self, method="ball_pivoting", *args, **kwargs):
#         logger.info("Preprocessing cloud of points")
#         assert self.is_data_ok(), "Data is not properly set-up"
#         self.point_cloud = o3d.geometry.PointCloud()
#         self.point_cloud.points = o3d.utility.Vector3dVector(self.data)
#         self.point_cloud.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=1.5,
#                                                           max_nn=75))
#         if config.stl_from_point_cloud.stl_strategy:
#             method = config.stl_from_point_cloud.stl_strategy
#         if method == 'ball_pivoting' or method == 'ball pivoting':
#             if 'ball_radius' in kwargs:
#                 self.run_ball_pivoting(ball_radius=kwargs['ball_radius'], **kwargs)
#             else:
#                 self.run_ball_pivoting()
#
#         if method == 'poisson':
#             self.run_poisson_reconstruction()
#
#     def is_data_ok(self):
#         if isinstance(self.data, pd.DataFrame):
#             self.preprocessed_data = self.data.values
#             return True
#         elif isinstance(self.data, np.ndarray):
#             self.preprocessed_data = self.data
#             return True
#         else:
#             return False
#
#     def visualize_point_cloud(self):
#         o3d.visualization.draw_geometries([self.point_cloud])
#
#     def visualize_stl_mesh(self, stl_mesh=None):
#         if stl_mesh is not None:
#             o3d.visualization.draw_geometries([stl_mesh])
#         else:
#             assert self.is_run, "Run the self.run() method before visualizing the results"
#             o3d.visualization.draw_geometries([self.stl_mesh])
#
#     @set_run
#     def run_ball_pivoting(self, ball_radius=None, simplify_mesh=False, **kwargs):
#         logger.info("Running the ball pivoting algorithm to recreate the stl surface")
#         if not ball_radius and not config.stl_from_point_cloud.ball_radius:
#             distances = self.point_cloud.compute_nearest_neighbor_distance()
#             self.avg_distance = np.mean(distances)
#             self.ball_radius = 3 * self.avg_distance * config.stl_from_point_cloud.buff if config.stl_from_point_cloud.buff else 1.0
#             self.stl_mesh: o3d.geometry.TriangleMesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(
#                 pcd=self.point_cloud,
#                 radii=o3d.utility.DoubleVector([self.ball_radius, self.ball_radius * 2]),
#                 **config.stl_from_point_cloud.arguments,
#             )
#         else:
#             self.ball_radius = ball_radius if ball_radius else config.stl_from_point_cloud.ball_radius
#             self.stl_mesh: o3d.geometry.TriangleMesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(
#                 pcd=self.point_cloud,
#                 radii=o3d.utility.DoubleVector([self.ball_radius, self.ball_radius * 2]),
#                 **config.stl_from_point_cloud.arguments,
#             )
#         if simplify_mesh or config.stl_from_point_cloud.simplify_mesh:
#             self.simplify_mesh()
#         return self.stl_mesh
#
#
#     def make_mesh_watertight(self,
#                              hole_size=None,
#                              filename=None,
#                              write=True):
#         """This method reads a ply mesh from an external file and converts it into a watertight mesh
#         """
#         self.to_ply(filename="temp.ply")
#         logger.info(f"Making the mesh read from {filename} watertight")
#         reader = vtk.vtkPLYReader()
#         if filename is None:
#             filename = Path.cwd() / "output/temp.ply"
#         reader.SetFileName(str(filename))
#         reader.Update()
#         polydata = reader.GetOutput()
#         fill = vtk.vtkFillHolesFilter()
#         fill.SetInputData(polydata)
#         hole_size = hole_size if hole_size else config.stl_from_point_cloud.watertight_mesh.hole_size if config.stl_from_point_cloud.watertight_mesh.hole_size else 1.0
#         fill.SetHoleSize(hole_size)
#         fill.Update()
#         self.vtk_filled = fill.GetOutput()
#         if write:
#             writer = vtk.vtkPLYWriter()
#             writer.SetInputData(self.vtk_filled)
#             writer.SetFileName(f"temp-watertight.ply")
#             writer.Write()
#         self.stl_mesh = o3d.io.read_triangle_mesh("temp-watertight.ply")
#         print(self.stl_mesh)
#         # Path.unlink(Path("temp-watertight.ply"))
#         # Path.unlink(filename)
#
#         return fill.GetOutput()
#
#     def reduce_mesh(self, target_value=100000):
#         self.stl_mesh = self.stl_mesh.reduce_mesh(target_value)
#         return self.stl_mesh
#
#
#     def simplify_mesh(self, target_value=100000):
#         self.stl_mesh.remove_degenerate_triangles()
#         self.stl_mesh.remove_duplicated_triangles()
#         self.stl_mesh.remove_duplicated_vertices()
#         self.stl_mesh.remove_non_manifold_edges()
#         return self.stl_mesh
#
#
#     @set_run
#     def run_poisson_reconstruction(self, depth=8, **kwargs):
#         logger.info("Running the poisson reconstruction method to recreate the stl surface")
#         self.point_cloud.orient_normals_towards_camera_location(self.point_cloud.get_center())
#         self.point_cloud.normals = o3d.utility.Vector3dVector(- np.asarray(self.point_cloud.normals))
#         self.stl_mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(self.point_cloud,
#                                                                                   depth=depth,
#                                                                                   **kwargs if kwargs else {},
#                                                                                   **config.stl_from_point_cloud.arguments,
#                                                                                   )[0]
#         return self.stl_mesh
#
#     def to_ply(self, filename="reconstructed_mesh.ply"):
#         """Writes the mesh in ply format
#         """
#         # assert self.is_run, "a mesh reconstruction technique needs to be run first"
#         o3d.io.write_triangle_mesh(str(self.output_directory / filename), self.stl_mesh)
#         logger.info(f"Writing to {str(self.output_directory / filename)}")
#
#     def to_stl(self, filename="reconstructed_mesh.stl"):
#         """Writes the mesh in stl format
#         """
#         # assert self.is_run, "a mesh reconstruction technique needs to be run first"
#         self.stl_mesh = self.stl_mesh.compute_vertex_normals()
#         logger.info(f"Writing to {str(self.output_directory / filename)}")
#         o3d.io.write_triangle_mesh(str(self.output_directory / filename), self.stl_mesh)
#
#     def read_stl(self, filename):
#         """Reads a stl file and converts it into a o3d mesh"""
#         self.stl_mesh = o3d.io.read_triangle_mesh(filename)
#
#     def crop_bounding_box(self, bbox=None):
#         """
#         Uses original cloud of point bounding box to crop the STL mesh
#         """
#         if bbox is None:
#             logger.info("Computing bounding box and cropping the mesh")
#             bbox = self.point_cloud.get_axis_aligned_bounding_box()
#         else:
#             logger.info(f"Computing bounding box at {bbox} and cropping the mesh")
#             bbox = o3d.geometry.AxisAlignedBoundingBox(min_bound=bbox[0], max_bound=bbox[1])
#
#         self.stl_mesh = self.stl_mesh.crop(bbox)
#         return self.stl_mesh
#
#     @staticmethod
#     def create_output_directory():
#         """Creates an ./output folder on the working directory
#         """
#         Path(Path.cwd() / "output").mkdir(parents=True, exist_ok=True)
#
#     @property
#     def output_directory(self):
#         self.create_output_directory()
#         return Path.cwd() / "output"
#



from .BaseManager import BaseManager
from . import PflotranStudy
import subprocess
import docker
from docker.models.containers import Container
from io import BytesIO
from pathlib import Path
from pydelling.utils import create_results_folder
from pydelling.managers.PflotranPostprocessing import PflotranPostprocessing
from pydelling.managers.ssh import JurecaSsh
from pydelling.managers.status import PflotranStatus
import logging
import os
import getpass
import os

logger = logging.getLogger(__name__)
class PflotranManager(BaseManager):
    def _get_study_status(self, study_id: int):
        """This method returns the status of a study.
        """
        pass

    def _run_study(self,
                   study: PflotranStudy,
                   n_cores: int = 1,
                   petsc_dir: str = '/opt/pflotran-dev/petsc',
                   petsc_arch: str = 'arch-linux-c-opt',
                   **kwargs,
                   ):
        """This method runs a study.
        """
        if n_cores == 1:
            # Run the study in serial
            subprocess.run(['pflotran', '-pflotranin', study.input_file_name], cwd=study.output_folder.absolute())
        else:
            # Run the study in parallel
            os.environ['PETSC_DIR'] = petsc_dir
            os.environ['PETSC_ARCH'] = petsc_arch
            subprocess.call([f'$PETSC_DIR/$PETSC_ARCH/bin/mpirun -np {n_cores} pflotran -pflotranin {study.input_file_name}'],
                           cwd=study.output_folder.absolute(),
                           shell=True
                           )

    def _run_study_docker(self,
                          study: PflotranStudy,
                          docker_image: str,
                          n_cores: int = 1,
                          **kwargs,
                          ):
        """This method runs a study using docker.
        """
        docker_client = docker.from_env()
            # Run the study in serial

        try:
            docker_client.api.start(self.manager_name)
            container = docker_client.containers.get(self.manager_name)
        except:
            docker_client.api.create_container(
                name=self.manager_name,
                image=docker_image,
                command='/bin/sh',
                detach=True,
                tty=True,
                volumes=[f'/home/pflotran/'],
                host_config=docker_client.api.create_host_config(
                    binds={
                        Path().cwd() / f'studies/{study.name}': {
                            'bind': f'/home/pflotran/{study.name}',
                            'mode': 'rw',
                        }
                    })
            )
            docker_client.api.start(container=self.manager_name)
            container = docker_client.containers.get(self.manager_name)

        # Compress study folder to a tar file
        bytes_io = BytesIO()
        import tarfile
        pw_tar = tarfile.TarFile(fileobj=bytes_io, mode='w')
        pw_tar.add(study.output_folder, arcname=study.output_folder.name)
        pw_tar.close()
        bytes_io.seek(0)
        # volume = docker_client.volumes.create(name=study.output_folder.name, driver='local', driver_opts={'type': 'none', 'device': study.output_folder, 'o': 'bind'})
        # Copy the tar file to the container
        container.exec_run(cmd="bash -c 'cd /home/pflotran'")
        container.put_archive('/home/pflotran', bytes_io)
        container.exec_run(cmd="bash -c 'ls'")
        if n_cores == 1:
            container.exec_run(cmd=f"sudo bash -c 'export PATH=$PATH:/home/pflotran/pflotran/src/pflotran &&"
                                   f" cd /home/pflotran/{study.output_folder.name} && "
                                   f"pflotran -pflotranin {study.input_file_name}'")
        else:
            result = container.exec_run(cmd=f"sudo bash -c 'export PATH=$PATH:/home/pflotran/pflotran/src/pflotran && "
                                   f"export PETSC_DIR=/home/pflotran/petsc && "
                                   f"export PETSC_ARCH=docker-petsc && "
                                   f"cd /home/pflotran/{study.output_folder.name} && "
                                   f"$PETSC_DIR/$PETSC_ARCH/bin/mpiexec -np {n_cores} pflotran -pflotranin {study.input_file_name}'")
        container.stop()
        container.remove()

    def _run_study_jureca(self,
                            study: PflotranStudy,
                            user,
                            project_name,
                            pkey_path,
                            n_cores: int = 1,
                            wallclock_limit: str = None,
                            shell_script_path: str = None,
                            **kwargs,
                            ):
        """This method runs a study on JURECA"""
        if self.password is None:
            self.password = getpass.getpass(prompt='Password: ', stream=None)

        self.ssh = JurecaSsh(user=user,
                             project_name=project_name,
                             pkey_path=pkey_path,
                             password=self.password,
                             )
        # Create a folder for the set of studies
        self.ssh.cd_project()
        self.ssh.mkdir(self.studies_folder_name)
        # Create a folder for the study, delete it if it already exists
        self.ssh.cd(self.studies_folder_name)
        if study.name in self.ssh.ls:
            self.ssh.rmdir(study.name)
        self.ssh.mkdir(study.name)
        # Copy the study folder to the server
        self.ssh.cpdir(study.output_folder, study.name)
        # Copy the shell_script to the server
        if shell_script_path is not None:
            # Read the shell script and replace the study name where {input_name} is found\
            with open(shell_script_path, 'r') as f:
                shell_script = f.read()
            shell_script = shell_script.replace('{input_name}', study.input_file_name)
            # Write to temporary file
            with open('temp.sh', 'w') as f:
                f.write(shell_script)
            # Delete the temporary file
            self.ssh.cp('temp.sh', f"{study.name}/{Path(shell_script_path).name}")
            os.remove('temp.sh')
            # Run the shell script
            self.ssh.send_job(f"{study.name}/{Path(shell_script_path).name}")
            # Get the highest job id
            job_id = max(self.ssh.user_queue['JOBID'])
            # pflotran_status = PflotranStatus()
            self.ssh.wait_for_job(job_id)
            # Copy the results back to the local machine
            # Detect files ending with .h5
            dir_list = self.ssh.ls
            dir_list = [file for file in dir_list if file.endswith('.h5')]
            # Add the files containing the job id to the list
            dir_list += [file for file in self.ssh.ls if str(job_id) in file]
            for file in dir_list:
                self.ssh.get(f"{file}", study.output_folder/file)

        else:
            raise ValueError('shell_script should be provided with the details of the job submission.')









    def merge_results(self, move=False, postprocess=True):
        """This method merges the results of all the studies.
        """
        self.run(dummy=True)
        import shutil
        # results_folder = create_results_folder(list(self.studies.values())[0].output_folder)
        self.merged_folder = create_results_folder(self.results_folder / 'merged_results')
        for study in self.studies.values():
            hdf5_files = list(study.output_folder.glob('*.h5'))
            for file in hdf5_files:
                if 'restart' not in file.name:
                    if not move:
                        shutil.copy(file, self.results_folder / 'merged_results')
                    else:
                        shutil.move(file, self.results_folder / 'merged_results')

        # Copy -domain.h5 file


        if postprocess:
            domain_folder = list(self.studies.values())[0].output_folder / 'input_files'
            domain_file = list(domain_folder.glob('*-domain.h5'))[0]
            shutil.copy(domain_file, self.results_folder / 'merged_results')
            logger.info('Postprocessing results')
            pflotran_postprocesser = PflotranPostprocessing()
            # Change the working directory to the results folder
            os.chdir(self.results_folder / 'merged_results')
            pflotran_postprocesser.run()
            # Return to the original working directory



















from pydelling.managers.ssh import BaseSsh
import paramiko
import getpass
import logging
import pandas as pd
from pathlib import Path
import rich
from pydelling.managers.status import BaseStatus
logger = logging.getLogger(__name__)


class JurecaSsh(BaseSsh):
    hostname = 'jureca.fz-juelich.de'
    def __init__(self,
                 user,
                 pkey_path,
                 project_name,
                 password=None,
                 ):
        """
        Connects to the remote server.
        Args:
            user: username
            pkey_path: path to the private key
        """
        super().__init__(user, pkey_path, password)
        self.project_name = project_name
        self.current_job_id = None

    def connect(self):
        # Ask for password using a hidden input
        if self.password is None:
            self.password = getpass.getpass(prompt='Password: ', stream=None)
        self.pkey = paramiko.ed25519key.Ed25519Key.from_private_key_file(self.pkey_path, password=self.password)
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.connect(
            hostname=self.hostname,
            username=self.user,
            password=self.password,
            pkey=self.pkey,
        )
        self.sftp = paramiko.SFTPClient.from_transport(self.client.get_transport())


    @property
    def general_queue(self) -> pd.DataFrame:
        """
        Returns the queue status in the remote server.
        Returns: queue status
        """
        queue_string = self.run_command("squeue")
        lines = queue_string.split('\n')
        lines = [line.split() for line in lines]
        # Take only the length of the first line
        lines = [line[:len(lines[0])] for line in lines]
        return pd.DataFrame(lines[1:], columns=lines[0])

    @property
    def user_queue(self) -> pd.DataFrame:
        """
        Returns the queue status in the remote server.
        Returns: queue status
        """
        queue_string = self.run_command("squeue -u aitor1")
        lines = queue_string.split('\n')
        lines = [line.split() for line in lines]
        # Take only the length of the first line
        lines = [line[:len(lines[0])] for line in lines]
        return pd.DataFrame(lines[1:], columns=lines[0])

    def cd_project(self):
        """
        Changes the current working directory to the project directory in the remote server.
        """
        self.cd(f'/p/scratch/cjiek63/{self.project_name}')

    def send_job(self, job_path):
        """
        Sends a job to the remote server.
        Args:
            job_path: path to the job file
        """
        job_path = Path(job_path)
        logger.info(f'Sending job {job_path} to the remote server')
        self.cd(job_path.parent)
        self.run_command(f"cd {self.pwd} && sbatch {job_path.name}")

    def wait_for_job(self,
                     job_id,
                     check_interval=3,
                     running_object_status: BaseStatus = None,
                     ):
        """
        Waits for a job to finish in the remote server.
        Args:
            job_id: id of the job
            check_interval: time in seconds between checks

        Returns:

        """
        import time
        from rich.progress import Progress

        logger.info(f'Waiting for job {job_id} to finish in the remote server')
        is_finished = False
        # Add a static rich status information bar (not a Progress object) that can be updated dynamically
        with Progress() as status_bar:
            task_wait = status_bar.add_task(f"Waiting the job {job_id} to start", total=None)
            task_start = status_bar.add_task(f"Starting the job {job_id}", total=None, visible=False)
            task_cancel = status_bar.add_task(f"Canceling the job {job_id}", total=None, visible=False)
            task_run = status_bar.add_task(f"Running the job {job_id}", total=None if running_object_status is None else 100, visible=False)
            assert_finish_counter = 0

            while not is_finished:
                job_status = self.get_job_status(job_id)
                if job_status == 'CF':
                    assert_finish_counter = 0
                    status_bar.update(task_wait, visible=False)
                    status_bar.update(task_cancel, visible=False)
                    status_bar.update(task_run, visible=False)
                    status_bar.update(task_start, visible=True)
                elif job_status == 'PD':
                    assert_finish_counter = 0
                    status_bar.update(task_wait, visible=True)
                    status_bar.update(task_cancel, visible=False)
                    status_bar.update(task_run, visible=False)
                    status_bar.update(task_start, visible=False)
                elif job_status == 'R':
                    assert_finish_counter = 0
                    status_bar.update(task_wait, visible=False)
                    status_bar.update(task_cancel, visible=False)
                    status_bar.update(task_run, visible=True)
                    status_bar.update(task_start, visible=False)
                    if running_object_status is not None:
                        running_object_status.read_status_file()
                        status_bar.update(task_run, completed=running_object_status.progress)
                elif job_status == 'CG':
                    status_bar.update(task_wait, visible=False)
                    status_bar.update(task_cancel, visible=True)
                    status_bar.update(task_run, visible=False)
                    status_bar.update(task_start, visible=False)
                elif job_status == None:
                    assert_finish_counter += 1
                    if assert_finish_counter > 3:
                        status_bar.remove_task(task_wait)
                        status_bar.remove_task(task_cancel)
                        status_bar.remove_task(task_run)
                        status_bar.remove_task(task_start)
                        is_finished = True

                time.sleep(check_interval)


    def cancel_job(self, job_id):
        """
        Cancels a job in the remote server.
        Args:
            job_id: id of the job
        """
        logger.info(f'Cancelling job {job_id} in the remote server')
        self.run_command(f"scancel {job_id}")

    def cancel_all_jobs(self):
        """
        Cancels all jobs in the remote server.
        """
        get_user_jobs = self.user_queue['JOBID']
        for job in get_user_jobs:
            self.cancel_job(job)

    def get_job_status(self, job_id):
        """
        Returns the status of a job in the remote server.
        Args:
            job_id: id of the job

        Returns: status of the job
        """
        try:
            return self.general_queue[self.general_queue['JOBID'] == job_id]['ST'].values[0]
        except:
            return None

    @property
    def user_job_ids(self):
        """
        Returns the jobs of the user in the remote server.
        Returns: jobs of the user
        """
        return self.user_queue['JOBID'].values







import logging

from ..preprocessing.mesh_preprocessor import MeshPreprocessor

logger = logging.getLogger(__name__)
from tqdm import tqdm
import meshio
from pathlib import Path
import numpy as np


class VTKMeshReader(MeshPreprocessor):
    has_kd_tree = False

    def __init__(self, filename,
                 kd_tree=True,
                 st_file=False,
                 generate_internal_mesh=True,
                 ):
        super().__init__()
        self.is_streamlit = st_file

        if Path(filename).suffix == '.vtk':
            self.meshio_mesh: meshio.Mesh = meshio.read(filename)
            self._coords = self.meshio_mesh.points
            if generate_internal_mesh:
                self.convert_meshio_to_meshpreprocessor()
                if kd_tree:
                    self.create_kd_tree()
                    self.has_kd_tree = True
        else:
            self.load(filename)


    def convert_meshio_to_meshpreprocessor(self):
        if self.is_streamlit:
            import streamlit as st
        for cell_block in self.meshio_mesh.cells:
            element_type = cell_block.type
            if element_type == "wedge":
                count = 0
                if self.is_streamlit:
                    st.write(f'Creating {len(cell_block.data)} wedge elements')
                    wedge_progress = st.empty()
                    wedge_progress.progress(0)
                for element in tqdm(cell_block.data, desc='Setting up wedge elements'):
                    if self.is_streamlit:
                        proportion = round(count / len(cell_block.data) * 100)
                        wedge_progress.progress(proportion)
                        count += 1

                    self.add_wedge(element, self._coords[element])

            elif element_type == 'hexahedron':
                count = 0
                if self.is_streamlit:
                    st.write(f'Creating {len(cell_block.data)} hexahedron elements')
                    hexahedron_progress = st.empty()
                    hexahedron_progress.progress(0)
                for element in tqdm(cell_block.data, desc='Setting up hexahedron elements'):
                    if self.is_streamlit:
                        proportion = round(count / len(cell_block.data) * 100)
                        hexahedron_progress.progress(proportion)
                        count += 1
                    self.add_hexahedra(element, self._coords[element])

            elif element_type == 'tetra':
                count = 0
                if self.is_streamlit:
                    st.write(f'Creating {len(cell_block.data)} tetrahedra elements')
                    tetra_progress = st.empty()
                    tetra_progress.progress(0)
                for element in tqdm(cell_block.data, desc='Setting up tetra elements'):
                    if self.is_streamlit:
                        proportion = round(count / len(cell_block.data) * 100)
                        tetra_progress.progress(proportion)
                        count += 1

                    self.add_tetrahedra(element, self._coords[element])

            elif element_type == 'pyramid':
                count = 0
                if self.is_streamlit:
                    st.write(f'Creating {len(cell_block.data)} pyramid elements')
                    pyramid_progress = st.empty()
                    pyramid_progress.progress(0)
                for element in tqdm(cell_block.data, desc='Setting up pyramid elements'):
                    if self.is_streamlit:
                        proportion = round(count / len(cell_block.data) * 100)
                        pyramid_progress.progress(proportion)
                        count += 1
                    self.add_pyramid(element, self._coords[element])


    def save(self, filename):
        """Save the mesh to a file."""
        import pickle
        logger.info(f'Saving mesh to {filename}')
        with open(filename, 'wb') as f:
            save_dictionary = {
                'elements': self.elements,
                'coords': self.coords,
                'kd_tree': self.kd_tree,
                'has_kd_tree': self.has_kd_tree,
            }
            pickle.dump(save_dictionary, f)

    def load(self, filename):
        """Load the mesh from a file."""
        logger.info(f'Loading mesh from {filename}')
        import pickle
        with open(filename, 'rb') as f:
            save_dictionary = pickle.load(f)
            self.elements = save_dictionary['elements']
            self._coords = save_dictionary['coords']
            self.kd_tree = save_dictionary['kd_tree']
            self.has_kd_tree = save_dictionary['has_kd_tree']

    @property
    def cell_data_values(self) -> dict:
        """Return the cell data from the meshio mesh. Data is structured for each cell type."""
        return self.meshio_mesh.cell_data_dict

    @property
    def point_data_values(self) -> dict:
        """Return the point data from the meshio mesh."""
        return self.meshio_mesh.point_data

    @property
    def cell_variables(self) -> list:
        """Return a list of cell variable names."""
        return list(self.meshio_mesh.cell_data.keys())

    @property
    def point_variables(self) -> list:
        """Return a list of point variable names."""
        return list(self.meshio_mesh.point_data.keys())

    @property
    def cell_data_flatten(self) -> dict:
        """Return the cell data from the meshio mesh. Data is flattened for each cell type into a single array."""
        data_dict = self.meshio_mesh.cell_data_dict
        temp_dict = {}
        for key, value in data_dict.items():
            temp_dict[key] = []
            for cell_type, cell_values in value.items():
                temp_dict[key] += cell_values.tolist()
            temp_dict[key] = np.array(temp_dict[key])
        return temp_dict









import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.neighbors import KernelDensity

from pydelling.utils.decorators import set_run

logger = logging.getLogger(__name__)
import seaborn as sns

class KdeEstimator:
    """
    This class performs a KDE estimation on a given dataset and provides useful methods
    to plot and manipulate the estimated distributions.
    """
    is_run: bool
    kde_estimator: KernelDensity
    def __init__(self, data:pd.DataFrame=None, kernel='gaussian', bandwidth=1000, package='scikit'):
        self.data = data
        self.kernel = kernel
        self.bandwidth = bandwidth
        self.package = package
        if type(self.data) == pd.Series:
            self.data = pd.DataFrame(self.data)

    @set_run
    def run(self, data: pd.DataFrame=None):
        """
        Fit the provided dataset to the corresponding kernel distribution
        """
        if data:
            self.data = data
        logger.info(f'Fitting KDE generator for {self.data.columns.to_list()} variables')
        self.kde_estimator: KernelDensity = KernelDensity(kernel=self.kernel, bandwidth=self.bandwidth).fit(self._training_data)
        return self.kde_estimator

    def plot_1d_comparison_histograms(self,
                                      variable=None,
                                      n=10000,
                                      savefig=None,
                                      bins=30,
                                      colors=None,
                                      xlabel=None,
                                      ylabel=None
                                      ):
        plt.clf()
        fig, ax = plt.subplots()
        fig: plt.Figure
        ax: plt.Axes
        if not variable:
            variable = self.data.columns[0]

        sampled_values = pd.DataFrame(self.kde_estimator.sample(n), columns=self.data.columns)

        ax.hist(self.data[variable].values, bins=bins,
                density=True,
                alpha=0.5,
                label=variable,
                color=colors[0] if colors else 'C0'
                )

        ax.hist(sampled_values[variable].values, bins=bins,
                density=True,
                alpha=0.5,
                label=f"{variable}-kde",
                color=colors[1] if colors else 'C1'
                )
        ax.set_axisbelow(True)
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)
        plt.legend()
        plt.grid()
        if savefig:
            plt.savefig(savefig)
        else:
            plt.show()
        return ax

    def plot_1d_comparison_histograms_multi(self, variables, n=10000, savefig=None, bins=30, palette=None, dpi=150):
        plt.clf()
        fig, ax = plt.subplots()
        fig: plt.Figure
        ax: plt.Axes

        sampled_values = pd.DataFrame(self.kde_estimator.sample(n), columns=self.data.columns)
        for idx, variable in enumerate(variables):
            if palette:
                color = palette[idx]
            sns.histplot(self.data[variable].values, bins=bins,
                    alpha=0.75,
                    label=variable,
                    color=color if palette else {}
                    )
            sns.histplot(sampled_values[variable].values, bins=bins,
                    alpha=0.25,
                    label=f"{variable}-kde",
                    color=color if palette else {}
                         )
        ax.set_axisbelow(True)
        plt.legend()
        plt.grid()
        if savefig:
            plt.savefig(savefig, dpi=dpi)
        else:
            plt.show()
        return ax

    def plot_1d(self, variable=None, n=100, savefig=None):
        """
        Plots the trained KDE using the original data bounds
        Returns: Axes object of the generated plot
        """
        plt.clf()
        if not variable:
            variable = self.data.columns[0]
        logger.info(f'Generating 1D comparison plot for {variable} variable')

        temp = [np.linspace(self.data[column].min(), self.data[column].max(), n) for column in self.data]
        sampling_data = np.meshgrid(*temp)
        sampling_data: np.ndarray = np.vstack(map(np.ravel, sampling_data)).T.reshape(-1, self.data.shape[1])

        predicted_distribution = self.kde_estimator.score_samples(sampling_data)
        predicted_distribution = np.exp(predicted_distribution)

        # Plot result
        fig, ax = plt.subplots()
        fig: plt.Figure
        ax: plt.Axes
        ax.hist(self.data[variable].values, bins=30,
                density=True,
                alpha=0.5,
                label=variable,
                )
        ax.plot(sampling_data,
                predicted_distribution,
                alpha=0.80,
                label=f'{variable}-predicted KDE'
                )
        plt.legend()
        plt.grid()
        ax.set_ylabel('Density [-]')
        ax.set_xlabel('Integration time [y]')
        if savefig:
            plt.savefig(savefig)
        else:
            plt.show()
        return ax

    @property
    def _training_data(self):
        assert not self.data.empty, 'Provide some data to work with'
        return self.data.values.reshape(-1, self.data.shape[1])

    def sample(self, *args, **kwargs):
        return self.kde_estimator.sample(*args, **kwargs)


from typing import Dict

import pandas as pd
from copy import deepcopy


class BaseReader:
    """Base class for the different reader subclasses"""
    filename: str
    data: Dict[str, pd.DataFrame]

    def read(self):
        """This function should read the data automatically from the config file"""
        pass

    def copy(self):
        """This function returns a copy of the reader"""
        return deepcopy(self)



import unittest
from pydelling.interpolation import BaseInterpolator
from pydelling.readers import iGPReader, CentroidReader
from pydelling.utils import test_data_path
import numpy


class BaseInterpolatorCase(unittest.TestCase):
    def setUp(self) -> None:
        self.mesh_data = iGPReader(test_data_path() / "test_implicit_to_explicit.gid", project_name="test").get_mesh()
        self.interpolation_data = CentroidReader(test_data_path() / "centroid_reader_data.csv", separator=",", header=True).get_data()
        self.base_interpolator = BaseInterpolator(mesh_data=self.mesh_data, interpolation_data=self.interpolation_data)

    def test_set_up(self):
        numpy.testing.assert_array_equal(self.base_interpolator.mesh, self.mesh_data)
        numpy.testing.assert_array_equal(self.base_interpolator.data, self.interpolation_data)

    def test_interpolation(self):
        interpolated_data = self.base_interpolator.run()
        numpy.testing.assert_array_equal(interpolated_data, self.mesh_data[:, 0])


if __name__ == '__main__':
    unittest.main()



from .BaseStudy import BaseStudy
from .PflotranPostprocessing import PflotranPostprocessing
from .PflotranStudy import PflotranStudy
from .BaseManager import BaseManager
from .PflotranManager import PflotranManager


from .callbacks import *
from .status import *


from abc import ABC, abstractmethod

from pydelling.webapps import WebAppRunner
from translate import Translator
import streamlit as st
import extra_streamlit_components as stx
from ..BaseStreamlitUtilityClass import BaseStreamlitUtilityClass

class BaseComponent(ABC, BaseStreamlitUtilityClass):
    _value = None
    def __init__(self,
                 webapp: WebAppRunner=None,
                 lang=None,
                 translate=False,
                 key=None,
                 *args,
                 **kwargs
                 ):
        self.type = self.__class__.__name__
        self.finished = False
        self.translate = translate

        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])

        if webapp:
            self.webapp = webapp
        if lang is not None:
            self.translator = Translator(to_lang=lang)
            self.translate = True
        self.name = f'{self.type}{key if key is not None else ""}'
        self.initialize_in_session_state(key=f'{self.name}-init', value=False)
        self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args, **kwargs):
        """Runs the component"""

    @property
    def value(self):
        """Returns the value of the component"""
        return self._value


from __future__ import annotations
from pathlib import Path
from jinja2 import Template
import shutil
from pydelling.utils import UnitConverter
from typing import Callable

from typing import TYPE_CHECKING, List, Union
if TYPE_CHECKING:
    from pydelling.managers import BaseCallback, BaseManager

import logging

logger = logging.getLogger(__name__)


class BaseStudy(UnitConverter):
    """This class contains the core methods for all the other child Manager classes.
    A manager is supposed to control the simulations run on a given software automatically.

    The manager should be able to:
    - Create the input files for the software
    - Modify the input files for the software
    - Run the software
    - Read the output status of the simnulation

    The idea to process the input files is the following:
    - Read the raw input file
    - The user can add variables to change, or can specify directly on the input file using {{var}} notation (jinja2)
    - The input file is rendered using jinja2 for each specific case
    """
    count = 0

    def __init__(self,
                 input_file: str,
                 study_name: str = None,
                 is_independent: bool = False,
                 input_file_name: str = None,
                 ):
        """This method initializes the manager.
        """
        if not is_independent:
            self.idx = self.__class__.count
            self.__class__.count += 1
            self.is_independent = False
        else:
            self.idx = None
            self.is_independent = True
        self.name = study_name if study_name is not None else f"{self.__class__.__name__}-{self.idx + 1}"
        self.input_file_name = input_file_name if input_file_name is not None else Path(input_file).name
        logger.info(f"Initializing {self.__class__.__name__} study")
        self.input_file = Path(input_file)
        self.settings = {}
        self.jinja_settings = {}
        # Step 1: read and set the input file
        self.raw_text = self.input_file.read_text()
        self.aux_files = {}
        self.output_folder = None
        self._callbacks: List[Callable] = []
        self.callbacks: List[BaseCallback] = []

    def pre_run(self):
        """This method is executed before the run.
        """
        pass

    def post_run(self):
        """This method is executed after the run.
        """
        pass

    def _replace_with_jinja_variable(self, var, var_name=None, value=None):
        """This method replaces a variable in the raw text with a jinja variable.
        """
        logger.debug(f"Replacing {var} with jinja variable")
        if var_name is None:
            var_name = var
        self.raw_text = self.raw_text.replace(var, f"{{{{ {var_name} }}}}")
        self.jinja_settings[var_name] = {"value": value}

    def _find_tags(self, tag: str, ignore_case: bool = True, equal: bool = True):
        """This method finds the line index of the tag in the raw text.
        """
        logger.debug(f"Finding tag '{tag}'")
        lines = self.raw_text.splitlines()
        if ignore_case:
            lines = [line.lower() for line in lines]
            tag = tag.lower()
        # Delete lines starting with #
        # lines = [line for line in lines if not line.startswith('#')]
        line_idx = [i for i, line in enumerate(lines) if tag in line]
        # if equal:
        #     lines_strip = [line.strip().split()[0] for line in lines]
        #     print(lines_strip)
        #     line_idx = [i for i in line_idx if lines[i].strip().strip()[0] == tag]
        #     print(line_idx)
        # Clean for comments
        line_idx = [i for i in line_idx if not lines[i].startswith('#')]
        return line_idx

    def _find_tags_in_subset(self, tags: list, subset: list, ignore_case: bool = True):
        """This method finds the line index of the tag in the raw text.
        """
        logger.debug(f"Finding tags '{tags}' in subset")
        if ignore_case:
            subset = [line.lower() for line in subset]
            tags = [tag.lower() for tag in tags]
        return [i for i, line in enumerate(subset) if any(tag in line for tag in tags)]

    def _get_line(self, line_index: int):
        """This method returns the line of the raw text.
        """
        return self.raw_text.splitlines()[line_index]

    def _add_line(self, line_index: int, new_line: list, sep: str = ' '):
        """This method adds a line in the raw text.
        """
        logger.debug(f"Adding line {line_index} with {new_line}")
        lines = self.raw_text.splitlines()
        lines.insert(line_index, sep.join(new_line))
        self.raw_text = '\n'.join(lines)

    def _get_nth_previous_line(self, line_index: int, n: int = 1):
        """This method returns the nth previous line of the raw text.
        """
        return self.raw_text.splitlines()[line_index - n]

    def _get_nth_next_line(self, line_index: int, n: int = 1):
        """This method returns the nth next line of the raw text.
        """
        return self.raw_text.splitlines()[line_index + n]

    def _replace_line(self, line_index: int, new_line: list, sep: str = ' '):
        """This method replaces a line in the raw text.
        """
        logger.debug(f"Replacing line {line_index} with {new_line}")
        lines = self.raw_text.splitlines()
        lines[line_index] = sep.join(new_line)
        self.raw_text = '\n'.join(lines)



    def render(self, **kwargs):
        """This method renders the input file using jinja2.
        """
        logger.info(f"Rendering input file {self.input_file_name}")
        template = Template(self.raw_text)
        return template.render(**kwargs)


    def add_auxiliary_file(self, file_path: str):
        """This method adds an auxiliary file to the manager.
        """
        logger.info(f"Adding auxiliary file {file_path}")
        file_path = Path(file_path)
        self.aux_files[file_path.name] = file_path

    def add_input_file(self, file_path: Union[Path, str]):
        """This method adds an auxiliary file to the manager.
        """
        logger.info(f"Adding input file {file_path}")
        file_path = Path(file_path)
        self.aux_files[file_path.name] = file_path

    def add_input_folder(self, folder_path: str):
        """This method adds an auxiliary folder to the manager.
        """
        folder_path = Path(folder_path)
        for file in folder_path.glob('*'):
            self.aux_files[file.name] = file
        logger.info(f"Adding input folder {folder_path} with {len(self.aux_files)} files")

    def add_callback(self, callback: Callable, kind: str = 'pre', **kwargs):
        """This method adds a callback to the manager.
        """
        kwargs['kind'] = kind
        self._callbacks.append(lambda manager: callback(manager, self, **kwargs))

    def initialize_callbacks(self, manager: BaseManager):
        """This method initializes the callbacks.
        """
        temp_callbacks = []
        for callback in self._callbacks:
            temp_callbacks.append(callback(manager))
        self.callbacks = temp_callbacks


    def to_file(self,
                output_folder: str=None,
                output_file: str=None,
                auxiliary_folder: str=None,
                **kwargs):
        """This method renders the input file and saves it to a file.
        """
        self.output_folder = output_folder if output_folder is not None else f'case-{BaseStudy.count}'
        logger.info(f"Saving input files to {output_folder}")
        BaseStudy.count += 1
        output_folder = Path(self.output_folder)
        output_folder.mkdir(exist_ok=True)
        output_file = output_folder / (output_file if output_file is not None else Path(self.input_file_name))
        output_file.write_text(self.render(**kwargs))

        # Copy the auxiliary files
        if len(self.aux_files) > 0:
            auxiliary_folder = auxiliary_folder if auxiliary_folder is not None else 'input_files'
            auxiliary_folder = output_folder / auxiliary_folder
            auxiliary_folder.mkdir(exist_ok=True)
            for aux_file in self.aux_files:
                shutil.copy(self.aux_files[aux_file], auxiliary_folder / aux_file)

    def to_tar(self):
        """This method creates a tar file with the input files. Using the gzip compression.
        """
        import tarfile
        import os
        logger.info(f"Creating tar file {self.output_folder}.tar")
        self.to_file(output_folder='temp')
        # Create tar file using gzip compression
        with tarfile.open(f"{self.output_folder}.tar", "w:gz") as tar:
            tar.add('temp', arcname=os.path.basename('temp'))
        shutil.rmtree('temp')

    def __repr__(self):
        return f"{self.__class__.__name__} (input template: {self.input_file.name})"

    # Properties
    def copy(self):
        """This method returns a copy of the object.
        """
        new_obj = self.__class__(self.input_file)
        for attr in self.__dict__:
            if attr in ['idx', 'name']:
                continue
            setattr(new_obj, attr, getattr(self, attr))
        return new_obj



    def __str__(self):
        return self.__repr__()















from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class SaveDataFilter(BaseFilter):
    """
    This class implements the Save Data filter
    """
    filter_type: str = "Save_data"
    counter: int = 0

    def __init__(self, filename, proxy, point_data_arrays, cell_data_arrays, name):
        super().__init__(name=name)
        SaveDataFilter.counter += 1
        self.filter = SaveData(self, Input=filename, proxy=proxy)
        self.filter.Proxy = proxy
        self.filter.PointDataArrays = point_data_arrays
        self.filter.CellDataArrays = cell_data_arrays



from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class TecplotFilter(BaseFilter):
    filter_type: str = "tecplot_reader"
    counter: int = 0

    def __init__(self, filename, name):
        super().__init__(name=name)
        self.filter = TecplotReader(FileNames=str(filename))
        TecplotFilter.counter += 1




from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class VtkFilter(BaseFilter):
    filter_type: str = "VTK_reader"
    counter: int = 0
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    z_min: float
    z_max: float

    def __init__(self, filename, name):
        super().__init__(name=name)
        self.filter = LegacyVTKReader(FileNames=str(filename))
        VtkFilter.counter += 1
        self.set_ranges()

    def set_ranges(self):
        self.x_min = self.mesh_points.min()["x"]
        self.x_max = self.mesh_points.max()["x"]
        self.y_min = self.mesh_points.min()["y"]
        self.y_max = self.mesh_points.max()["y"]
        self.z_min = self.mesh_points.min()["z"]
        self.z_max = self.mesh_points.max()["z"]


import numpy as np
import pandas as pd
import h5py

class PflotranProcessingUtils:
    """This class contains utility functions to use on the PflotranReader class"""
    variables: list
    coordinates: np.ndarray
    data: h5py.File
    # data:
    axis_translator = {
        'x': "x[m]",
        'y': "y[m]",
        'z': "z[m]",
    }

    def get_slice(self, data: np.ndarray, axis: str, index: int) -> np.ndarray:
        """This function returns a slice of the data array along the axis and index provided"""
        assert len(data.shape) == 3, "The full data array must be provided"
        assert axis in ['x', 'y', 'z'], "The axis must be x, y, or z"

        if axis == 'x':
            return data[index, :, :]
        elif axis == 'y':
            return data[:, index, :]
        elif axis == 'z':
            return data[:, :, index]

    def get_slice_from_coordinates(self, data: np.ndarray, axis: str, coordinate: float) -> np.ndarray:
        """This function returns a slice of the data array along the axis and index provided"""
        assert len(data.shape) == 3, "The full data array must be provided"
        assert axis in ['x', 'y', 'z'], "The axis must be x, y, or z"

        axis_full = self.axis_translator[axis]
        # Find closes index of coordinate
        index = np.argmin(np.abs(self.coordinates[axis_full] - coordinate))

        if axis == 'x':
            sliced_data = data[index, :, :]
            # Expand the dimension of the axis to match the data
            sliced_data = np.expand_dims(sliced_data, axis=0)
            return sliced_data

        elif axis == 'y':
            sliced_data = data[:, index, :]
            # Expand the dimension of the axis to match the data
            sliced_data = np.expand_dims(sliced_data, axis=1)
            return sliced_data

        elif axis == 'z':
            sliced_data = data[:, :, index]
            # Expand the dimension of the axis to match the data
            sliced_data = np.expand_dims(sliced_data, axis=2)
            return sliced_data


    def get_shape_dimensions(self, data: np.ndarray) -> tuple:
        """This function returns the cartesians dimensions present on a data array"""
        dims = ''
        if data.shape[0] > 1:
            dims += 'x'
        if data.shape[1] > 1:
            dims += 'y'
        if data.shape[2] > 1:
            dims += 'z'
        return dims


    def axis_centroids(self, axis):
        """This function returns the centroids of the axis provided"""
        assert axis in ['x', 'y', 'z'], "The axis must be x, y, or z"
        axis = self.axis_translator[axis]
        return np.diff(self.coordinates[axis]) + self.coordinates[axis][0:-1]
    @property
    def x_centroid(self):
        return np.diff(self.coordinates['x[m]']) + self.coordinates['x[m]'][0:-1]

    @property
    def y_centroid(self):
        return np.diff(self.coordinates['y[m]']) + self.coordinates['y[m]'][0:-1]

    @property
    def z_centroid(self):
        return np.diff(self.coordinates['z[m]']) + self.coordinates['z[m]'][0:-1]

    @property
    def x_min(self):
        return self.coordinates['x[m]'][0]

    @property
    def x_max(self):
        return self.coordinates['x[m]'][-1]

    @property
    def y_min(self):
        return self.coordinates['y[m]'][0]

    @property
    def y_max(self):
        return self.coordinates['y[m]'][-1]

    @property
    def z_min(self):
        return self.coordinates['z[m]'][0]

    @property
    def z_max(self):
        return self.coordinates['z[m]'][-1]

    @property
    def x_extent(self):
        return self.x_max - self.x_min

    @property
    def y_extent(self):
        return self.y_max - self.y_min

    @property
    def z_extent(self):
        return self.z_max - self.z_min

    @property
    def x_spacing(self):
        return np.diff(self.coordinates['x[m]'])

    @property
    def y_spacing(self):
        return np.diff(self.coordinates['y[m]'])

    @property
    def z_spacing(self):
        return np.diff(self.coordinates['z[m]'])






from .BaseStudy import BaseStudy
from abc import ABC, abstractmethod
from typing import Dict, List, Union
import logging
from alive_progress import alive_bar
from tqdm import tqdm
from pydelling.utils import create_results_folder
from pydelling.managers.ssh import BaseSsh
from docker import DockerClient
from pathlib import Path
import subprocess


logger = logging.getLogger(__name__)


class BaseManager(ABC):
    """This class is the base class for all simulation managers. A simulation manager should be able to:
    - Manage study [BaseStudy] objects
    - Run simulations on a given software
    - Read the output status of the simulation
    """
    def __init__(self, name: str = None):
        self.results_folder = None
        self.studies: Dict[str, BaseStudy] = {}
        self.manager_name = name if name is not None else self.__class__.__name__
        self.is_dummy = False
        self.ssh: BaseSsh = None
        self.password = None


    def add_study(self, study: BaseStudy):
        """This method adds a study to the manager.
        """
        assert isinstance(study, BaseStudy), f"Study must be a object from a class inherited from BaseStudy, not {type(study)}"
        self.studies[study.name] = study

    def run(self,
            studies_folder: str = './studies',
            n_cores: int = 1,
            docker_image: str = None,
            dummy: bool = False,
            start_from: int = None,
            petsc_dir: str = '/opt/pflotran-dev/petsc',
            petsc_arch: str = 'arch-linux-c-opt',
            pre_commands: List[str] = None,
            run_on_jureca: bool = False,
            user: str = None,
            project_name: str = None,
            pkey_path: str = None,
            wallclock_limit: float = 23,
            shell_script: str = None,
            **kwargs,
            ):
        """This method runs all the studies.
        """
        self.is_dummy = dummy
        shell_script = Path(shell_script).absolute()
        # Initialize callbacks
        for study in self.studies.values():
            study.initialize_callbacks(self)

        self.results_folder = create_results_folder(studies_folder)
        self.studies_folder_name = Path(studies_folder).name
        # self.generate_run_files(studies_folder=studies_folder)
        # Run the precommands if any
        if pre_commands is not None:
            for command in pre_commands:
                logger.info(f"Running precommand: {command}")
                subprocess.run(command, shell=True)
        for study in tqdm(self.studies.values(), desc="Running studies", colour="white"):
            study: BaseStudy
            if start_from is not None:
                if study.idx < start_from - 1:
                    logger.info(
                        f"Skipping study {study.name} (idx: {study.idx}) because start_from is set to {start_from}")
                    self.run_study(study,
                                   docker_image=docker_image,
                                   n_cores=n_cores,
                                   dummy=True,
                                   petsc_dir=petsc_dir,
                                   petsc_arch=petsc_arch,
                                   run_on_jureca=run_on_jureca,
                                   user=user,
                                   project_name=project_name,
                                   pkey_path=pkey_path,
                                   wallclock_limit=wallclock_limit,
                                   shell_script_path=shell_script,
                                   **kwargs)
                    continue
            self.run_study(study,
                           docker_image=docker_image,
                           n_cores=n_cores,
                           dummy=dummy,
                           petsc_dir=petsc_dir,
                           petsc_arch=petsc_arch,
                           run_on_jureca=run_on_jureca,
                           user=user,
                           project_name=project_name,
                           pkey_path=pkey_path,
                           wallclock_limit=wallclock_limit,
                           shell_script_path=shell_script,
                           **kwargs)

    def generate_run_files(self, studies_folder: str = './studies'):
        """This method generates the run files for all the studies.
        """
        for study in self.studies.values():
            study: BaseStudy
            study.to_file(self.results_folder / study.name)

    @abstractmethod
    def _run_study(self, study: BaseStudy, n_cores: int = 1, **kwargs):
        """This method runs a study.
        """
        logger.info(f"Running study {study.name}")
        return NotImplementedError("This method must be implemented in the child class")

    @abstractmethod
    def _run_study_docker(self,
                          study: BaseStudy,
                          docker_image: str,
                          n_cores: int = 1,
                          **kwargs,
                          ):
        """This method runs a study using docker.
        """
        logger.info(f"Running study {study.name} using docker image {docker_image}")
        return NotImplementedError("This method must be implemented in the child class")

    def _run_study_jureca(self,
                            study: BaseStudy,
                            user: str,
                            project_name: str,
                            pkey_path: str,
                            n_cores: int = 1,
                            wallclock_limit: float = None,
                            shell_script: str = None,
                            **kwargs,
                            ):
        """This method runs a study in JURECA."""
        logger.info(f"Running study {study.name} in JURECA")
        return NotImplementedError("This method must be implemented in the child class")

    def run_study(self,
                  study: BaseStudy,
                  n_cores: int = 1,
                  docker_image: str = None,
                  run_on_jureca: bool = False,
                  dummy: bool = False,
                  user: str = None,
                  project_name: str = None,
                  pkey_path: str = None,
                  wallclock_limit: float = 23,
                  shell_script: str = None,
                  **kwargs,
                  ):
        """This method runs a study.
        """
        logger.info(f"Running study {study.name}")
        # Create the study files
        study.output_folder = self.results_folder / study.name
        if dummy:
            logger.info("Dummy run, not running the study")
        else:
            for callback in study.callbacks:
                if callback.kind == 'pre':
                    callback.run()
            study.to_file(self.results_folder / study.name)
            # Run the study
            if run_on_jureca:
                self._run_study_jureca(study,
                                       n_cores=n_cores,
                                       user=user,
                                       project_name=project_name,
                                       pkey_path=pkey_path,
                                       wallclock_limit=wallclock_limit,
                                       shell_script=shell_script,
                                       **kwargs)
            elif docker_image is not None:
                self._run_study_docker(study, docker_image, n_cores=n_cores, **kwargs)
            else:
                self._run_study(study, n_cores=n_cores, **kwargs)
            for callback in study.callbacks:
                if callback.kind == 'post':
                    callback.run()
            study.post_run()


    @property
    def n_studies(self):
        """This method returns the number of studies.
        """
        return len(self.studies)

    def merge_studies(self):
        """This methods merges the studies generated by the manager.
        """
        pass








from WebAppRunner import WebAppRunner


class TestWebApp(WebAppRunner):
    def construct(self):
        import streamlit.components.v1 as components
        import pyvista
        from pyvista import examples
        mesh = examples.load_uniform()
        pl = pyvista.Plotter(shape=(1, 2))
        _ = pl.add_mesh(mesh, scalars='Spatial Point Data', show_edges=True)
        pl.subplot(0, 1)
        _ = pl.add_mesh(mesh, scalars='Spatial Cell Data', show_edges=True)
        pl.export_html('pyvista.html')  # doctest:+SKIP

        HtmlFile = open("pyvista.html", 'r', encoding='utf-8')
        source_code = HtmlFile.read()
        components.html(source_code, height=500, width=500)



if __name__ == '__main__':
    test_webapp = TestWebApp()
    test_webapp.run()



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydelling.managers import PflotranManager, PflotranStudy

from pydelling.managers.callbacks.BaseCallback import BaseCallback
from pydelling.managers import PflotranPostprocessing
from pathlib import Path
import shutil
import logging

logger = logging.getLogger(__name__)


class PflotranSaveResultsCallback(BaseCallback):
    """Callback to restart Pflotran simulations."""
    move: bool = False
    postprocess: bool = False
    def __init__(self, manager: PflotranManager, study: PflotranStudy, kind: str = 'post', **kwargs):
        super().__init__(manager, study, 'post', **kwargs)

    def run(self):
        """This callback runs after the simulation is run, it creates a folder and copies (or moves) the results to it"""
        move = self.kwargs['move'] if 'move' in self.kwargs else False
        postprocess = self.kwargs['postprocess'] if 'postprocess' in self.kwargs else False

        output_files = list(self.study.output_folder.glob('*h5'))
        target_folder = Path(self.manager.results_folder / 'merged_results')
        target_folder.mkdir(exist_ok=True, parents=True)
        for file in output_files:
            if 'restart' in file.name:
                continue
            if file.stem[-1] == 'y':
                continue
            if move:
                # Check if the file already exists
                if (target_folder / file.name).exists():
                    target_file = target_folder / file.name
                    target_file.unlink()
                shutil.move(str(file), str(target_folder.absolute()))
            else:
                shutil.copy(str(file), str(target_folder.absolute()))

        if postprocess:
            import os
            domain_folder = list(self.manager.studies.values())[0].output_folder / 'input_files'
            domain_file = list(domain_folder.glob('*-domain.h5'))[0]
            shutil.copy(domain_file, self.manager.results_folder / 'merged_results')
            old_cwd = os.getcwd()
            logger.info('Postprocessing results')
            pflotran_postprocesser = PflotranPostprocessing()
            # Change the working directory to the results folder
            os.chdir(self.manager.results_folder / 'merged_results')
            pflotran_postprocesser.run()
            os.chdir(old_cwd)







import numpy as np

from pydelling.preprocessing.mesh_preprocessor.geometry import BaseFace


class QuadrilateralFace(BaseFace):
    def __init__(self, node_ids, node_coords):
        super().__init__(node_ids, node_coords)
        self.type = "quadrilateral"

    def compute_centroid(self):
        return np.mean(self.coords, axis=0)
        # t1 = [0, 1, 3]
        # t2 = [1, 2, 3]
        # triangles = [t1, t2]
        # poly_centroid = np.zeros(shape=3)
        # for triangle_nodes in triangles:
        #     # Set-up variables
        #     q1 = self.node_coords[triangle_nodes[0]]  # Point1 of small triangle
        #     q2 = self.node_coords[triangle_nodes[1]]  # Point2 of small triangle
        #     q3 = self.node_coords[triangle_nodes[2]]  # Point3 of small triangle
        #     q_v = np.array([q1, q2, q3])
        #     v1 = q2 - q1
        #     v2 = q3 - q1
        #     v_cross = np.cross(v1, v2)
        #     area_triangle = np.linalg.norm(v_cross) / 2.0
        #     # print(f"area triangle:{area_triangle}")
        #     # print(f"Total area:{self.area}")
        #     # Compute centroid
        #     mean_centroid = np.mean(q_v, axis=0)  # Centroid of small triangle
        #     poly_centroid += area_triangle * mean_centroid
        #     # print(f"q1: {q1}, q2: {q2}, q3:{q3}, mean_centroid:{mean_centroid}")
        # return poly_centroid / self.area

    @property
    def edges(self):
        return [
            [self.nodes[0], self.nodes[1]],
            [self.nodes[1], self.nodes[2]],
            [self.nodes[2], self.nodes[3]],
            [self.nodes[3], self.nodes[0]]
        ]

    @property
    def edge_vectors(self):
        return [
            self.coords[1] - self.coords[0],
            self.coords[2] - self.coords[1],
            self.coords[3] - self.coords[2],
            self.coords[0] - self.coords[3],
        ]



from typing import List

import numpy as np

from pydelling.utils.geometry_utils import order_points_clockwise
from . import Point, Segment
from .BasePrimitive import BasePrimitive


class Polygon(BasePrimitive):
    def __init__(self, points: List[Point] or np.ndarray):
        # The set of points should be ordered in a clockwise fashion
        self.points = order_points_clockwise(points)
        self.segments = self.generate_segments()

    def generate_segments(self):
        segments = []
        for i in range(len(self.points)):
            segments.append(Segment(self.points[i], self.points[(i + 1) % len(self.points)]))
        return segments

    def to_csv(self, filename='polygon.csv'):
        import csv
        with open(filename, 'w') as csvfile:
            writer = csv.writer(csvfile, delimiter=',')
            for point in self.points:
                writer.writerow(point)

                






import numpy as np

from pydelling.preprocessing.mesh_preprocessor.geometry import BaseFace


class TriangleFace(BaseFace):
    def __init__(self, node_ids, node_coords):
        super().__init__(node_ids, node_coords)
        self.type = "triangle"

    def compute_centroid(self):
        return np.mean(self.coords, axis=0)

    @property
    def edges(self):
        return [
            [self.nodes[0], self.nodes[1]],
            [self.nodes[1], self.nodes[2]],
            [self.nodes[2], self.nodes[0]]
        ]

    @property
    def edge_vectors(self):
        return [
            self.coords[1] - self.coords[0],
            self.coords[2] - self.coords[1],
            self.coords[0] - self.coords[2]
        ]



"""
Author: Aitor Iraola
Date: 030920
Description: This script postprocesses a given PFLOTRAN simulation carried out using an unstructured explicit mesh and the methodology developed by Amphos 21
"""
from pathlib import Path
import h5py
import os
import shutil
import numpy as np
from typing import Dict
import logging

logger = logging.getLogger(__name__)

class PflotranPostprocessing:
    """This class manages the structure of the output files"""
    def __init__(self, dt=0.2, unit="y"):
        self.dt = dt
        self.unit = unit
        self.output_times = []
        self.output_units = []
        self.project_name = None

    def run(self):
        "Method that runs the post-processing engine"
        self.find_directories()
        self.find_output_files()
        self.find_domain_file()
        self.find_input_file()
        if len(self.output_vtk_vel_files) != 0:
            self.add_velocity_to_hdf5()
        self.find_attributes()
        self.export_xmf()
        self.copy_domain_file()

    def find_output_files(self):
        """This method searches the current directory and the directory called "./output-hdf5" to find PFLOTRAN output files"""
        # First, see if the directory contains the output files
        if self.current_directory.stem == "output-hdf5":
            self.output_h5_files = list(self.current_directory.glob("*.h5"))
            self.output_directory = self.current_directory
            assert len(self.output_h5_files) > 0, "Output hdf5 files couldn't be found"
        else:
            self.output_h5_files = list(self.root_directory.glob("*.h5"))
            self.output_directory = self.current_directory
            if len(self.output_h5_files) == 0:
                # Check the output-hdf5 folder
                output_folder = self.current_directory / "output-hdf5"
                self.output_h5_files = list(output_folder.glob("*.h5"))
                # assert len(self.output_h5_files) > 0, "Output hdf5 files couldn't be found"
                self.output_directory = self.current_directory / "output-hdf5"
        # Take out the domain hdf5 file from the output file list
        # Set the project name
        domain_file = [file for file in self.output_h5_files if "domain" in file.name]
        assert len(domain_file) > 0, "Domain file couldn't be found"
        # Remove the domain files from the list
        for file in domain_file:
            self.output_h5_files.remove(file)
        restart_file = [file for file in self.output_h5_files if "restart" in file.name]
        if len(restart_file) > 0:
            self.output_h5_files.remove(restart_file[0])
        self.project_name = self.output_h5_files[0].stem.split('-')[0]
        # Rename domain file using the project name
        domain_file[0].rename(self.output_directory / f"{self.project_name}-domain.h5")
        self.output_h5_files = sorted(self.output_h5_files)
        # Process the VTK files
        if self.current_directory.stem == "output-vtk":
            self.output_vtk_vel_files = list(self.current_directory.glob("*vel*.vtk"))
            self.output_vtk_directory = self.current_directory
            assert len(self.output_vtk_vel_files) > 0, "Output hdf5 files couldn't be found"
        else:
            self.output_vtk_vel_files = list(self.root_directory.glob("*vel*.vtk"))
            self.output_vtk_directory = self.current_directory
            if len(self.output_vtk_vel_files) == 0:
                # Check the output-hdf5 folder
                output_folder = self.current_directory / "output-vtk"
                self.output_vtk_vel_files = list(output_folder.glob("*vel*.vtk"))
                # assert len(self.output_vtk_files) > 0, "Output vtk files couldn't be found"
                self.output_vtk_directory = self.current_directory / "output-vtk"
        self.output_vtk_vel_files = sorted(self.output_vtk_vel_files)

    def find_directories(self):
        """Test """
        self.current_directory = Path.cwd()
        if self.current_directory.stem == "output-hdf5":
            self.root_directory = self.current_directory.parent
        elif self.current_directory.glob("*.in"):
            self.root_directory = self.current_directory
        elif self.current_directory.stem == "input_files":
            self.root_directory = self.current_directory.parent
        else:
            raise FileNotFoundError("Root directory couldn't be found")
        self.input_directory = self.root_directory / "input_files"

    def find_domain_file(self):
        """This method finds the domain file"""
        if len(list(self.input_directory.glob("*-domain.h5"))) > 0:
            self.domain_file = list(self.input_directory.glob("*-domain.h5"))[0]
        else:
            self.domain_file = list(self.current_directory.glob("*-domain.h5"))[0]
        return self.domain_file

    def find_input_file(self):
        self.input_file = None
        try:
            self.input_file = list(self.input_directory.glob("*.in"))[0]
        except:
            logger.warning("Input file couldn't be found")
        return self.input_file

    def add_velocity_to_hdf5(self):
        for i in range(0, self.n_output_files):
            with h5py.File(self.output_h5_files[i], "r+") as h5_file:
                vtk_filename = self.output_vtk_vel_files[i]
                vtk_file_dict = self.read_vtk_file(filename=vtk_filename)
                for var in vtk_file_dict:
                    if var == "Material_ID":
                        continue
                    hdf5_group = list(h5_file.values())[0]
                    var_names = [name for name in hdf5_group]
                    if not var in var_names:
                        hdf5_group.create_dataset(name=var, data=vtk_file_dict[var]["data"])

    def find_attributes(self):
        # Find number of cells and number of vertices
        with h5py.File(self.domain_file, "r") as file:
            self.n_vertices = len(file["Domain"]["Vertices"])
            self.pflotran_vertices = len(file["Domain"]["Cells"])
        # Find output variables
        with h5py.File(self.output_h5_files[0], "r") as hdf5_file:
            hdf5_group = list(hdf5_file.values())[0]
            self.output_variables = [item[0] for item in hdf5_group.items()]
            self.n_cells = len(hdf5_group[self.output_variables[0]])
        for hdf5_file in self.output_h5_files:
            if 'domain' in hdf5_file.name:
                # Delete the domain file from the list
                self.output_h5_files.remove(hdf5_file)
                continue

            with h5py.File(hdf5_file, "r") as hdf5_file:
                attribute_list = list(hdf5_file.items())[0][0].split()
                self.output_times.append(float(attribute_list[2]))
                self.output_units.append(attribute_list[3])

    def copy_domain_file(self):
        # check if file exists and delete otherwise
        # if list(self.output_directory.glob("*-domain.h5")):
        #     os.remove(list(self.output_directory.glob("*-domain.h5"))[0])
        # target_file = self.output_directory / "{}-domain.h5".format(self.project_name)
        # domain_file = str(self.domain_file)
        # shutil.copy(domain_file, str(target_file))
        pass

    def export_xmf(self):
        # Input variables
        dt = self.dt  # time interval between individual result files
        file_name = self.project_name
        Nelements = self.n_cells
        Nvertices = self.pflotran_vertices - self.n_cells
        Nvertices_pflotran = self.n_vertices


        # Script to generate the files
        for i in range(0, self.n_output_files):
            self.name = "{}-{:03d}.xmf".format(file_name, i)
            self.time = dt * i
            name_h5 = "{}-{:03d}.h5".format(file_name, i)
            self.export_file = self.output_directory
            xdmf_file = self.output_directory / self.name
            with open(str(xdmf_file), 'w') as self.output_file:
                self.output_file.write('<?xml version="1.0" ?>\n')
                self.output_file.write('<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n')
                self.output_file.write('<Xdmf>\n')
                self.output_file.write('\t<Domain>\n')
                self.output_file.write('\t<Grid Name="Mesh">\n')
                self.output_file.write('\t\t<Time Value = "{:1.5E}" />\n'.format(self.output_times[i]))
                self.output_file.write('\t\t<Topology Type="Mixed" NumberOfElements="%s" >\n' % Nelements)
                self.output_file.write('\t\t\t<DataItem Format="HDF" DataType="Int" Dimensions="%s">\n' % (Nelements + Nvertices))
                self.output_file.write('\t\t\t\t %s-domain.h5:/Domain/Cells\n' % file_name)
                self.output_file.write('\t\t\t</DataItem>\n')
                self.output_file.write('\t\t</Topology>\n')
                self.output_file.write('\t\t<Geometry GeometryType="XYZ">\n')
                self.output_file.write('\t\t\t<DataItem Format="HDF" Dimensions="%s 3">\n' % Nvertices_pflotran)
                self.output_file.write('\t\t\t\t %s-domain.h5:/Domain/Vertices\n' % file_name)
                self.output_file.write('\t\t\t</DataItem>\n')
                self.output_file.write('\t\t</Geometry>\n')
                for output_variable in self.output_variables:
                    self.export_attribute(output_variable, i)
                self.output_file.write('\t</Grid>\n')
                self.output_file.write('\t</Domain>\n')
                self.output_file.write('</Xdmf>\n')

    def export_attribute(self, var, i):
        self.output_file.write('\t\t<Attribute Name="%s" AttributeType="Scalar"  Center="Cell">\n' % var)
        self.output_file.write('\t\t\t<DataItem Dimensions="%s 1" Format="HDF">\n' % self.n_cells)
        self.output_file.write('\t\t\t\t{}-{:03d}.h5:/{:4d} Time  {:1.5E} {}/{}\n'.format(self.project_name, i, i, self.output_times[i], self.output_units[i], var))
        self.output_file.write('\t\t\t</DataItem>\n')
        self.output_file.write('\t\t</Attribute>\n')

    @staticmethod
    def read_vtk_file(filename) -> Dict:
        """This method reads an VTK file and returns the data"""
        # Process vtk file
        print("Processing {} VTK file".format(filename))
        data_dict = {}
        temp_array = []
        start_reading = False
        # filename = "G:/My Drive/dev/2020-project_repos/rwm_repo/scripts/3-post_processing/B-VTK_velocity/output-vtk/test-trial.vtk"
        with open(filename, "r") as vtk_file:
            for line in vtk_file.readlines():
                split_line = line.split()
                if not split_line:
                    start_reading = False
                if "CELL_DATA" in split_line:
                    n_cells = split_line[1]
                if "SCALARS" in split_line:
                    data_dict[split_line[1]] = {"name": split_line[1],
                                                "n_cells": n_cells,
                                                "data": []
                                                }
                    try:
                        data_dict[current_reading]["data"] = np.concatenate(temp_array)
                        data_dict[current_reading]["shape"] = data_dict[current_reading]["data"].shape
                    except:
                        pass
                    start_reading = False
                    current_reading = split_line[1]
                    print("Starting to read variable {}".format(current_reading))
                if "LOOKUP_TABLE" in split_line:
                    start_reading = True
                    temp_array = []
                    continue
                if start_reading:
                    temp_array.append(np.array(split_line).astype(np.float))
            data_dict[current_reading]["data"] = np.concatenate(temp_array)
            data_dict[current_reading]["shape"] = np.concatenate(temp_array).shape
        return data_dict

    @property
    def n_output_files(self):

        return len(self.output_h5_files)



import logging
import os
from pathlib import Path

import h5py
import numpy as np

from .BaseWriter import BaseWriter

logger = logging.getLogger(__name__)

class HDF5CentroidWriter(BaseWriter):
    def run(self, filename=None, remove_if_exists=True, include_cell_id=True):
        if filename is not None:
            self.filename = filename
        if remove_if_exists:
            try:
                os.remove(self.filename)
            except FileNotFoundError as ef:
                print("Nothing to overwrite!")

        if self.check_data():
            if not os.path.exists(self.filename):
                h5temp = h5py.File(self.filename, "w")
                h5temp.close()
            with h5py.File(self.filename, "r+") as h5temp:
                h5temp.create_dataset(self.var_name, data=self.data)
                if include_cell_id:
                    cell_id = np.array([index + 1 for index in range(len(self.data))])
                    h5temp.create_dataset("Cell Ids", data=cell_id)
        else:
            print("Couldn't find data to dump!")

    def write_anisotropic_dataset(self,
                                  dataset_x,
                                  dataset_y,
                                  dataset_z,
                                  filename=None,
                                  remove_if_exists=True,
                                  var_name=None,
                                  ):
        """
        Writes anisotropic dataset to HDF5
        Args:
            filename: name of the file
            dataset_x: dataset containing data in the x-direction
            dataset_y: dataset containing data in the y-direction
            dataset_z: dataset containing data in the z-direction
            remove_if_exists: removes file if exists
        """
        filename_path = Path(filename if filename else self.filename)
        logger.info(f"Writing anisotropic permeability to {filename_path}")
        if remove_if_exists:
            if filename_path.exists():
                filename_path.unlink()
        with h5py.File(filename_path, "w") as h5_temp:
            var_name_write = var_name if var_name else self.var_name
            h5_temp.create_dataset(f"{var_name_write}X", data=dataset_x)
            h5_temp.create_dataset(f"{var_name_write}Y", data=dataset_y)
            h5_temp.create_dataset(f"{var_name_write}Z", data=dataset_z)
            cell_id = np.array([index + 1 for index in range(len(dataset_x))])
            h5_temp.create_dataset("Cell Ids", data=cell_id)

    def write_full_tensor_dataset(self,
                                  dataset_x,
                                  dataset_xy,
                                  dataset_xz,
                                  dataset_y,
                                  dataset_yz,
                                  dataset_z,
                                  filename=None,
                                  remove_if_exists=True,
                                  var_name=None,
                                  ):
        """
        Writes anisotropic dataset to HDF5
        Args:
            filename: name of the file
            dataset_x: dataset containing data in the x-direction
            dataset_xy: dataset containing data in the xy-direction
            dataset_xz: dataset containing data in the xz-direction
            dataset_y: dataset containing data in the y-direction
            dataset_yz: dataset containing data in the yz-direction
            dataset_z: dataset containing data in the z-direction
            remove_if_exists: removes file if exists
        """
        filename_path = Path(filename if filename else self.filename)
        logger.info(f"Writing full tensor permeability to {filename_path}")
        if remove_if_exists:
            if filename_path.exists():
                filename_path.unlink()
        with h5py.File(filename_path, "w") as h5_temp:
            var_name_write = var_name if var_name else self.var_name
            h5_temp.create_dataset(f"{var_name_write}X", data=dataset_x)
            h5_temp.create_dataset(f"{var_name_write}XY", data=dataset_xy)
            h5_temp.create_dataset(f"{var_name_write}XZ", data=dataset_xz)
            h5_temp.create_dataset(f"{var_name_write}Y", data=dataset_y)
            h5_temp.create_dataset(f"{var_name_write}YZ", data=dataset_yz)
            h5_temp.create_dataset(f"{var_name_write}Z", data=dataset_z)
            cell_id = np.array([index + 1 for index in range(len(dataset_x))])
            h5_temp.create_dataset("Cell Ids", data=cell_id)


    def write_dataset(self,
                      dataset,
                      filename=None,
                      remove_if_exists=True,
                      var_name=None,
                      ):
        """
        Writes isotropic dataset to HDF5
        Args:
            filename: name of the file
            dataset_x: output dataset
            remove_if_exists: removes file if exists
        """
        filename_path = Path(filename if filename else self.filename)
        logger.info(f"Writing anisotropic permeability to {filename_path}")
        if remove_if_exists:
            if filename_path.exists():
                filename_path.unlink()
        with h5py.File(filename_path, "w") as h5_temp:
            var_name_write = var_name if var_name else self.var_name
            h5_temp.create_dataset(f"{var_name_write}", data=dataset)
            cell_id = np.array([index + 1 for index in range(len(dataset))])
            h5_temp.create_dataset("Cell Ids", data=cell_id)


from pydelling.webapps import WebAppRunner
import streamlit as st

class PflotranRunnerWebapp(WebAppRunner):
    """This webapp allows to run PFLOTRAN simulations on JURECA and visualize the evolution"""
    def construct(self):
        txt = st.text_area('Text to analyze', '''
             It was the best of times, it was the worst of times, it was
             the age of wisdom, it was the age of foolishness, it was
             the epoch of belief, it was the epoch of incredulity, it
             was the season of Light, it was the season of Darkness, it
             was the spring of hope, it was the winter of despair, (...)
             ''', height=1000, disabled=True)


if __name__ == '__main__':
    webapp = PflotranRunnerWebapp()
    webapp.run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydelling',
 'pydelling.config',
 'pydelling.interpolation',
 'pydelling.managers',
 'pydelling.managers.callbacks',
 'pydelling.managers.ssh',
 'pydelling.managers.status',
 'pydelling.paraview_processor',
 'pydelling.paraview_processor.filters',
 'pydelling.preprocessing',
 'pydelling.preprocessing.dfn_preprocessor',
 'pydelling.preprocessing.gid_preprocessor',
 'pydelling.preprocessing.mesh_preprocessor',
 'pydelling.preprocessing.mesh_preprocessor.geometry',
 'pydelling.readers',
 'pydelling.readers.iGPReader',
 'pydelling.readers.iGPReader.engines',
 'pydelling.readers.iGPReader.geometry',
 'pydelling.readers.iGPReader.io',
 'pydelling.readers.iGPReader.utils',
 'pydelling.tests',
 'pydelling.tests.interpolation',
 'pydelling.tests.managers',
 'pydelling.tests.managers.status',
 'pydelling.tests.preprocessing',
 'pydelling.tests.preprocessing.DfnPreprocessor',
 'pydelling.tests.preprocessing.MeshPreprocessor',
 'pydelling.tests.readers',
 'pydelling.tests.utils',
 'pydelling.tests.utils.geometry',
 'pydelling.tests.writers',
 'pydelling.utils',
 'pydelling.utils.geometry',
 'pydelling.webapps',
 'pydelling.webapps.components',
 'pydelling.webapps.webapps',
 'pydelling.writers']

package_data = \
{'': ['*'],
 'pydelling.tests': ['_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/custom_edge_intersections.txt',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/element.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/fracture.obj',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/gid_batch.bch',
                     '_trial_temp/studies/BaseStudy-2/*',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test.json',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_dfn.vtk',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_element.obj',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     '_trial_temp/test_mesh.vtk',
                     'case-5/*',
                     'case-6/*',
                     'results/Pwall.ex',
                     'results/Pwall.ex',
                     'results/test-implicit_to_explicit.mesh',
                     'results/test-implicit_to_explicit.mesh',
                     'studies/BaseStudy-2/*',
                     'test_data/*',
                     'test_data/problematic_fractures/*',
                     'test_data/test_implicit_to_explicit.gid/*',
                     'test_data/test_implicit_to_explicit.gid/PFLOTRAN/*',
                     'test_data/test_implicit_to_explicit.gid/PFLOTRAN/PFLOTRANrun/*',
                     'test_folder/*'],
 'pydelling.tests.managers': ['_trial_temp/studies/BaseStudy-2/*'],
 'pydelling.tests.readers': ['_trial_temp/*'],
 'pydelling.tests.utils.geometry': ['_trial_temp/*'],
 'pydelling.webapps': ['.streamlit/*']}

install_requires = \
['alive-progress',
 'dill',
 'docker',
 'h5py',
 'matplotlib',
 'meshio',
 'mkdocs',
 'mkdocs-material',
 'mkdocstrings',
 'mpmath',
 'natsort',
 'numpy',
 'numpy-stl',
 'ofpp',
 'openpyxl',
 'pandas',
 'paramiko',
 'plotly',
 'python-box',
 'pyyaml',
 'rich',
 'rtree',
 'scikit-learn',
 'scipy',
 'seaborn',
 'setuptools',
 'shapely',
 'sklearn',
 'streamlit',
 'tabulate',
 'tqdm',
 'trimesh',
 'vtk',
 'xlrd']

setup_kwargs = {
    'name': 'pydelling',
    'version': '1.11',
    'description': 'Package providing a variety of utility methods for mathematical modelling',
    'long_description': ' # Pydelling\nSet of pre- and post-processing scripts for modelling\n\n\n## Recommendations\n\nUsing and IDE like PyCharm is recommended to keep things easy. It may be useful if you need to debug your scripts,\nor edit them.\n\nUsing PyCharm, you can follow the installation step and then run your scripts in pydelling env automatically.\nOtherwise, you can follow the steps below to run pydelling manually. \n\n\n## Installation\nConfigure the conda environment running\n\n`conda env create --file environment.yaml`\n\nRegular pip dependencies can be found at `requirements.txt`\n\n\n## Templates and Examples\n\nSome examples of usage of pydelling may be found inside `templates` folder. Use `config.yaml` to setup data sources and \nother variables.\n\n\n## Running template scripts\n\nIn order to execute pydelling, `pydelling` conda env needs to be loaded.\n\n\n### Loading conda environment\n\n`conda activate pydelling`\n\n### Running script\n\nRun the script and set the config file as argument. If the config file is not specified, `./config.yaml` will be taken as default.\n\nExample: \n\n`python templates\\interpolate_top_BC\\interpolate_BC.py templates\\interpolate_top_BC\\config.yaml`\n\n',
    'author': 'Aitor',
    'author_email': 'aitirga@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
